from __future__ import annotations

import datetime
import logging
import subprocess
import sys
import threading
from pathlib import PurePath
from types import TracebackType
from typing import IO, Any, Callable, Generator, List, Optional, TextIO, Type, Union

from .._docstring import no_example
from .._utils import random_port

__all__ = (
    "ShinyAppProc",
    "run_shiny_app",
    # For internal use only
    # "shiny_app_gen",
)


class OutputStream:
    """
    Designed to wrap an IO[str] and accumulate the output using a bg thread

    Also allows for blocking waits for particular lines.
    """

    def __init__(self, io: IO[str], desc: Optional[str] = None):
        self._io = io
        self._closed = False
        self._lines: List[str] = []
        self._cond = threading.Condition()
        self._thread = threading.Thread(
            group=None, target=self._run, daemon=True, name=desc
        )

        self._thread.start()

    def _run(self):
        """
        Add lines into self._lines in a tight loop.
        """

        try:
            while not self._io.closed:
                try:
                    line = self._io.readline()
                except ValueError:
                    # This is raised when the stream is closed
                    break
                if line != "":
                    with self._cond:
                        self._lines.append(line)
                        self._cond.notify_all()
        finally:
            # If we got here, we're finished reading self._io and need to signal any
            # waiters that we're done and they'll never hear from us again.
            with self._cond:
                self._closed = True
                self._cond.notify_all()

    def wait_for(self, predicate: Callable[[str], bool], timeout_secs: float) -> bool:
        """
        Wait until the predicate returns True for a line in the output.

        Parameters
        ----------
        predicate
            A function that takes a line of output and returns True if the line
            satisfies the condition.
        timeoutSecs
            How long to wait for the predicate to return True before raising a
            TimeoutError.
        """
        timeout_at = datetime.datetime.now() + datetime.timedelta(seconds=timeout_secs)
        pos = 0
        with self._cond:
            while True:
                while pos < len(self._lines):
                    if predicate(self._lines[pos]):
                        return True
                    pos += 1
                if self._closed:
                    return False
                else:
                    remaining = (timeout_at - datetime.datetime.now()).total_seconds()
                    if remaining < 0 or not self._cond.wait(timeout=remaining):
                        # Timed out
                        raise TimeoutError(
                            "Timeout while waiting for Shiny app to become ready"
                        )

    def __str__(self):
        with self._cond:
            return "".join(self._lines)


def dummyio() -> TextIO:
    io = TextIO()
    io.close()
    return io


class ShinyAppProc:
    """
    Class that represents a running Shiny app process.

    This class is a context manager that can be used to run a Shiny app in a subprocess. It provides a way to interact
    with the app and terminate it when it is no longer needed.
    """

    file: PurePath
    """The path to the Shiny app file."""
    proc: subprocess.Popen[str]
    """The subprocess object that represents the running Shiny app."""
    port: int
    """The port that the Shiny app is running on."""
    url: str
    """The URL that the Shiny app is running on."""
    stdout: OutputStream
    """The standard output stream of the Shiny app subprocess."""
    stderr: OutputStream
    """The standard error stream of the Shiny app subprocess."""

    def __init__(
        self,
        proc: subprocess.Popen[str],
        port: int,
        *,
        app_file: PurePath | str,
    ):
        self.proc = proc
        self.port = port
        self.url = f"http://127.0.0.1:{port}/"
        self.stdout = OutputStream(proc.stdout or dummyio())
        self.stderr = OutputStream(proc.stderr or dummyio())
        threading.Thread(group=None, target=self._run, daemon=True).start()

        self.file = PurePath(app_file)

    def _run(self) -> None:
        self.proc.wait()
        if self.proc.stdout is not None:
            self.proc.stdout.close()
        if self.proc.stderr is not None:
            self.proc.stderr.close()

    def close(self) -> None:
        """
        Closes the connection and terminates the process.

        This method is responsible for closing the connection and terminating the process associated with it.
        """
        # from time import sleep
        # sleep(0.5)
        self.proc.terminate()

    def __enter__(self) -> ShinyAppProc:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ):
        self.close()

    def wait_until_ready(self, timeout_secs: float) -> None:
        """
        Waits until the shiny app is ready to serve requests.

        Parameters
        ----------
        timeout_secs
            The maximum number of seconds to wait for the app to become ready.

        Raises
        ------
        ConnectionError
            If there is an error while starting the shiny app.
        TimeoutError
            If the shiny app does not become ready within the specified timeout.
        """
        error_lines: List[str] = []

        def stderr_uvicorn(line: str) -> bool:
            error_lines.append(line)
            if "error while attempting to bind on address" in line:
                raise ConnectionError(f"Error while starting shiny app: `{line}`")
            return "Uvicorn running on" in line

        if self.stderr.wait_for(stderr_uvicorn, timeout_secs=timeout_secs):
            return
        else:
            raise TimeoutError(
                "Shiny app exited without ever becoming ready. Waiting for 'Uvicorn running on' in stderr. Last 20 lines of stderr:\n"
                + "\n".join(error_lines[-20:])
            )


@no_example()
def run_shiny_app(
    app_file: Union[str, PurePath],
    *,
    start_attempts: int = 3,
    port: int = 0,
    cwd: Optional[str] = None,
    wait_for_start: bool = True,
    timeout_secs: float = 30,
    bufsize: int = 64 * 1024,
) -> ShinyAppProc:
    """
    Run a Shiny app in a subprocess.

    Parameters
    ----------
    app_file
        The path to the Shiny app file.
    start_attempts
        Number of attempts to try and start the Shiny app. If the random port is already
        in use, a new random port will be chosen and another attempt will be made. If
        all attempts have been made, an error will be raised.
    port
        The port to run the app on. If 0, a random port will be chosen.
    cwd
        The working directory to run the app in.
    wait_for_start
        If True, wait for the app to become ready before returning.
    timeout_secs
        The maximum number of seconds to wait for the app to become ready.
    bufsize
        The buffer size to use for stdout and stderr.

    Returns
    -------
    :
        A :class:`shiny.run.ShinyAppProc` object representing the running
        Shiny app process.
    """
    shiny_port = port if port != 0 else random_port()

    child = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "shiny",
            "run",
            "--port",
            str(shiny_port),
            str(app_file),
        ],
        bufsize=bufsize,
        executable=sys.executable,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd,
        encoding="utf-8",
    )

    # TODO: Detect early exit

    sa = ShinyAppProc(child, shiny_port, app_file=app_file)

    if wait_for_start:
        try:
            sa.wait_until_ready(timeout_secs)
        except ConnectionError as e:
            logging.error(f"Failed to bind to port: {e}")

            # Make sure the current process is closed
            sa.close()

            start_attempts -= 1
            if start_attempts < 1:
                # Ran out of attempts!
                raise e

            # Try again with a new port!
            return run_shiny_app(
                app_file,
                start_attempts=start_attempts,
                port=port,
                cwd=cwd,
                wait_for_start=wait_for_start,
                timeout_secs=timeout_secs,
                bufsize=bufsize,
            )

    return sa


# Internal method to help make fixtures a little easier to write
# Attempt up to 3 times to start the app, with a random port each time
def shiny_app_gen(
    app_file: PurePath | str,
    *,
    start_attempts: int = 3,
    port: int = 0,
    cwd: Optional[str] = None,
    # wait_for_start: bool = False,
    timeout_secs: float = 30,
    bufsize: int = 64 * 1024,
) -> Generator[ShinyAppProc, Any, None]:
    """
    Run a Shiny app in a subprocess.

    This app will be automatically shut down when the Generator is exhausted. A
    generator is returned so we can utilize the context manager methods of the
    `ShinyAppProc` class (`__enter__` and `__exit__`). This allows for the app to be
    automatically shut down when the context manager exists. (This exit method is not
    possible when returning a ShinyAppProc directly.)

    Parameters
    ----------
    app
        The path to the Shiny app file.
    start_attempts
        Number of attempts to try and start the Shiny app. If the random port is already
        in use, a new random port will be chosen and another attempt will be made. If
        all attempts have been made, an error will be raised.
    port
        The port to run the app on. If 0, a random port will be chosen.
    cwd
        The working directory to run the app in.
    timeout_secs
        The maximum number of seconds to wait for the app to become ready.
    bufsize
        The buffer size to use for stdout and stderr.

    Yields
    ------
    :
        A single Shiny app process
    """
    # wait_for_start
    #     If True, wait for the app to become ready before returning.

    sa = run_shiny_app(
        app_file,
        wait_for_start=True,
        start_attempts=start_attempts,
        port=port,
        cwd=cwd,
        bufsize=bufsize,
        timeout_secs=timeout_secs,
    )
    had_connection_error: bool = False
    try:
        with sa:
            yield sa
    except ConnectionError as e:
        had_connection_error = True
        raise e
    finally:
        if not had_connection_error:
            logging.warning("Application output:\n" + str(sa.stderr))



import re

from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_text")


def test_input_text_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    # page.set_default_timeout(1000)

    caption = controller.InputText(page, "caption")
    caption.expect.to_have_value("Data summary")
    expect(caption.loc).to_have_value("Data summary")

    expect(caption.loc_label).to_have_text("Caption:")

    # Better approach
    expect(caption.loc_label).to_have_text("Caption:")
    expect(caption.loc).to_have_value("Data summary")
    expect(caption.loc).not_to_have_attribute("width", re.compile(r".*"))
    expect(caption.loc).not_to_have_attribute("placeholder", re.compile(r".*"))
    expect(caption.loc).to_have_attribute("autocomplete", "off")
    expect(caption.loc).not_to_have_attribute("spellcheck", re.compile(r".*"))

    # Best approach
    caption.expect_label("Caption:")
    caption.expect_value("Data summary")
    caption.expect_width(None)
    caption.expect_placeholder(None)
    caption.expect_autocomplete("off")
    caption.expect_spellcheck(None)


def test_input_text_typical(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    caption = controller.InputText(page, "caption")
    caption.expect.to_have_value("Data summary")
    caption.set("test value")
    caption.expect.not_to_have_value("Data summary")
    caption.expect.to_have_value("test value")


def test_input_text_app(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    caption = controller.InputText(page, "caption")
    caption.expect.to_have_value("Data summary")

    value = controller.OutputTextVerbatim(page, "value")
    value.expect_value("Data summary")

    caption.set("test value")
    caption.expect.to_have_value("test value")
    value.expect_value("test value")



import os
from pathlib import Path
from typing import Any

app_dir = Path(__file__).parent
env_file = app_dir / ".env"


def load_dotenv(dotenv_path: os.PathLike[str] = env_file, **kwargs: Any) -> None:
    """
    A convenience wrapper around `dotenv.load_dotenv` that warns if `dotenv` is not installed.
    It also returns `None` to make it easier to ignore the return value.
    """
    try:
        import dotenv

        dotenv.load_dotenv(dotenv_path=dotenv_path, **kwargs)
    except ImportError:
        import warnings

        warnings.warn(
            "Could not import `dotenv`. If you want to use `.env` files to "
            "load environment variables, please install it using "
            "`pip install python-dotenv`.",
            stacklevel=2,
        )



from shiny.express import ui

# Set some Shiny page options
ui.page_opts(
    title="Hello Shiny Chat",
    fillable=True,
    fillable_mobile=True,
)

# Create a welcome message
welcome = ui.markdown(
    """
    Hi! This is a simple Shiny `Chat` UI. Enter a message below and I will
    simply repeat it back to you. For more examples, see this
    [folder of examples](https://github.com/posit-dev/py-shiny/tree/main/examples/chat).
    """
)

# Create a chat instance
chat = ui.Chat(
    id="chat",
    messages=[welcome],
)

# Display it
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get the user's input
    user = chat.user_input()
    # Append a response to the chat
    await chat.append_message(f"You said: {user}")



from shiny import render
from shiny.express import input, ui

ui.page_opts(full_width=True)

with ui.card(id="card"):
    ui.input_slider("a", "A", 1, 100, 50)

    @render.code
    def txt():
        return input.a()



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_sidebar(
    ui.sidebar(ui.input_slider("controller", "Controller", min=1, max=3, value=1)),
    ui.navset_card_tab(
        ui.nav_panel("Panel 1", "Panel 1 content", value="panel1"),
        ui.nav_panel("Panel 2", "Panel 2 content", value="panel2"),
        ui.nav_panel("Panel 3", "Panel 3 content", value="panel3"),
        id="inTabset",
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    def _():
        ui.update_navs("inTabset", selected="panel" + str(input.controller()))


app = App(app_ui, server)



__all__ = ("markdown",)

import importlib
import textwrap
import warnings
from typing import Callable, Literal, Optional

from .._docstring import add_example
from . import HTML


@add_example()
def markdown(
    text: str, *, render_func: Optional[Callable[[str], str]] = None, **kwargs: object
) -> HTML:
    """
    Convert a string of markdown to :func:`ui.HTML`.

    Parameters
    ----------
    text
        A string of text containing markdown.
    render_func
        A function (with at least 1 argument) which accepts a string of markdown and
        returns a string of HTML. By default, a customized instance of the
        `markdown_id.main.MarkdownIt` class (which supports Github-flavored markdown) from the
        ``markdown-it`` package is used.
    **kwargs
        Additional keyword arguments passed to the ``render_func``.

    Returns
    -------
    :
        An :class:`~shiny.ui.HTML` string of the rendered markdown.
    """

    if render_func is None:
        render_func = default_md_renderer()

    html = render_func(textwrap.dedent(text), **kwargs)

    return HTML(html)


# -----------------------------------------------------------------------------
# TODO: someday it might make sense for the default markdown parser to have support for
# more stuff like syntax highlighting, mathjax, etc., but two things aren't yet entirely
# clear to me:
#   1. Whether it makes more sense to use Quarto or MyST-Parser.
#   2. How to discover and properly include the relevant HTML dependencies
# -----------------------------------------------------------------------------


def default_md_renderer(
    preset: Literal["commonmark", "gfm"] = "gfm"
) -> Callable[[str], str]:
    try:
        from markdown_it.main import MarkdownIt
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "The default markdown parser requires the markdown-it-py package"
            " to be installed. Install it with `pip install markdown-it`."
        )

    if preset == "commonmark":
        parser = MarkdownIt("commonmark")
    else:
        try:
            importlib.import_module(name=".", package="linkify_it")
        except ModuleNotFoundError:
            warnings.warn(
                "The 'autolinking' feature of GitHub flavored markdown requires the "
                "linkify-it package. Install it with `pip install linkify-it`.",
                stacklevel=2,
            )
        # Inspired by MyST-Parser's gfm-only option
        # https://github.com/executablebooks/MyST-Parser/blob/ce1245b25/myst_parser/main.py#L257-L269
        parser = MarkdownIt("commonmark", {"linkify": True})
        parser.enable(["table", "linkify", "strikethrough"])

        try:
            from mdit_py_plugins.tasklists import tasklists_plugin

            parser.use(tasklists_plugin)  # type: ignore
        except ModuleNotFoundError:
            warnings.warn(
                "The 'tasklists' feature of GitHub flavored markdown requires the "
                "mdit_py_plugins package. Install it with `pip install mdit_py_plugins`.",
                stacklevel=2,
            )

    def _render(text: str) -> str:
        return parser.render(text)  # type: ignore

    return _render



import matplotlib.pyplot as plt
import numpy as np

from shiny import render
from shiny.express import input, ui

ui.input_slider("n", "N", 1, 100, 50)


@render.plot
def histogram():
    np.random.seed(19680801)
    x = 100 + 15 * np.random.randn(437)
    plt.hist(x, input.n(), density=True)



import pathlib

import pandas as pd

from shiny.express import input, render, ui

dir = pathlib.Path(__file__).parent
mtcars = pd.read_csv(dir / "mtcars.csv")


ui.input_checkbox("highlight", "Highlight min/max values")


@render.table
def result():
    if not input.highlight():
        # If we're not highlighting values, we can simply
        # return the pandas data frame as-is; @render.table
        # will call .to_html() on it.
        return mtcars
    else:
        # We need to use the pandas Styler API. The default
        # formatting options for Styler are not the same as
        # DataFrame.to_html(), so we set a few options to
        # make them match.
        return (
            mtcars.style.set_table_attributes(
                'class="dataframe shiny-table table w-auto"'
            )
            .hide(axis="index")
            .format(
                {
                    "mpg": "{0:0.1f}",
                    "disp": "{0:0.1f}",
                    "drat": "{0:0.2f}",
                    "wt": "{0:0.3f}",
                    "qsec": "{0:0.2f}",
                }
            )
            .set_table_styles([dict(selector="th", props=[("text-align", "right")])])
            .highlight_min(color="silver")
            .highlight_max(color="yellow")
        )


# Legend
with ui.panel_conditional("input.highlight"):
    with ui.panel_absolute(bottom="6px", right="6px", class_="p-1 bg-light border"):
        "Yellow is maximum, grey is minimum"



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.layout_column_wrap(
        ui.input_radio_buttons(
            "pet_type", "Pet type", ["Dog", "Cat", "Bird"], inline=True
        ),
        ui.input_radio_buttons("pet_sex", "Pet sex", ["Male", "Female"], inline=True),
        ui.input_text("name", "Pet name", "Charlie"),
        ui.input_text("royal_name", "Royal Name", "King Charlie"),
        width=1 / 2,
    )
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.pet_type)
    def _():
        # Update the label of the pet name input
        ui.update_text("name", label=f"{input.pet_type()}'s name")

    @reactive.effect
    def _():
        # Update the value of the royal name input
        royal_noun = "King" if input.pet_sex() == "Male" else "Queen"
        ui.update_text("royal_name", value=f"{royal_noun} {input.name()}")


app = App(app_ui, server)



from shiny.express import input, render, ui

ui.page_opts(title="Kitchen Sink: ui.input_text_area()", fillable=True)

with ui.layout_columns():
    with ui.card():
        ui.input_text_area("default", label="Default text area")

        @render.code
        def default_txt():
            return str(input.default())

    with ui.card():
        ui.input_text_area(
            "placeholder", "With placeholder", placeholder="Enter text here"
        )

        @render.code
        def placeholder_txt():
            return str(input.placeholder())


with ui.layout_columns():
    with ui.card():
        ui.input_text_area(
            "custom_size",
            "Custom size",
            width="300px",
            height="150px",
            value="Resized text area",
        )

        @render.code
        def custom_size_txt():
            return str(input.custom_size())

    with ui.card():
        ui.input_text_area(
            "rows",
            "Custom rows",
            rows=5,
            value="This text area has 5 rows",
            resize="none",
        )

        @render.code
        def rows_txt():
            return str(input.rows())

    with ui.card():
        ui.input_text_area(
            "cols", "Custom cols", cols=30, value="This text area has 30 cols"
        )

        @render.code
        def cols_txt():
            return str(input.cols())


with ui.layout_columns():
    with ui.card():
        ui.input_text_area(
            "autocomplete", "Autocomplete", autocomplete="name", resize="horizontal"
        )

        @render.code
        def autocomplete_txt():
            return str(input.autocomplete())

    with ui.card():
        ui.input_text_area(
            "resize", "Resizable", value="You can resize this text area", resize="both"
        )

        @render.code
        def resize_txt():
            return str(input.resize())

    with ui.card():
        ui.input_text_area(
            "spellcheck",
            "Spellcheck",
            value="paticular",
            spellcheck="true",
            resize="vertical",
        )

        @render.code
        def spellcheck_txt():
            return str(input.spellcheck())



from __future__ import annotations

from queue import PriorityQueue
from typing import Generic, TypeVar

T = TypeVar("T")


class PriorityQueueFIFO(Generic[T]):
    """
    Similar to queue.PriorityQueue, except that if two elements have the same
    priority, they are returned in the order they were inserted. Also, the item
    is kept separate from the priority value (with PriorityQueue, the priority
    is part of the item).
    """

    def __init__(self) -> None:
        # Using Tuple instead of tuple because in Python 3.8 and earlier, tuple isn't
        # generic
        self._pq: PriorityQueue[tuple[int, int, T]] = PriorityQueue()
        self._counter: int = 0

    def put(self, priority: int, item: T) -> None:
        """
        Add an item to the queue.

        Parameters
        ----------
        priority
            The priority of the item. Higher priority items will come out of the queue
            before lower priority items.

        item
            The item to put in the queue.
        """
        self._counter += 1
        self._pq.put((-priority, self._counter, item))

    def get(self) -> T:
        iteminfo: tuple[int, int, T] = self._pq.get()
        return iteminfo[2]

    def empty(self) -> bool:
        return self._pq.empty()



from __future__ import annotations

import json
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time
import warnings
from typing import Any, Callable, List, Optional, TypeVar

import pytest
import requests
from conftest import ScopeName

from shiny.run._run import shiny_app_gen

is_interactive = hasattr(sys, "ps1")
reruns = 1 if is_interactive else 3
reruns_delay = 1

LOCAL_LOCATION = "local"

__all__ = (
    "local_deploys_app_url_fixture",
    "skip_if_not_chrome",
)

# connect
server_url = os.environ.get("DEPLOY_CONNECT_SERVER_URL")
api_key = os.environ.get("DEPLOY_CONNECT_SERVER_API_KEY")
# shinyapps.io
shinyappsio_name = os.environ.get("DEPLOY_SHINYAPPS_NAME")
shinyappsio_token = os.environ.get("DEPLOY_SHINYAPPS_TOKEN")
shinyappsio_secret = os.environ.get("DEPLOY_SHINYAPPS_SECRET")

run_on_ci = os.environ.get("CI", "False") == "true"
repo = os.environ.get("GITHUB_REPOSITORY", "unknown")

should_use_github_requirements_txt = (
    os.environ.get("DEPLOY_GITHUB_REQUIREMENTS_TXT", "true").lower() == "true"
)


deploy_locations: List[str] = ["connect", "shinyapps"]


CallableT = TypeVar("CallableT", bound=Callable[..., Any])


def skip_if_not_chrome(fn: CallableT) -> CallableT:
    # # Keeping commented to allow for easier local debugging
    # import platform
    # fn = pytest.mark.skipif(
    #     platform.python_version_tuple()[:2] != ("3", "10"),
    #     reason="Test requires Python 3.10",
    # )(fn)
    fn = pytest.mark.only_browser("chromium")(fn)

    return fn


def skip_on_webkit(fn: CallableT) -> CallableT:
    fn = pytest.mark.skip_browser("webkit")(fn)

    return fn


def skip_on_python_version(
    version: str,
    reason: Optional[str] = None,
) -> Callable[[CallableT], CallableT]:

    reason_str = reason or f"Do not run on python {version}"

    is_valid_version = (
        re.match(r"\d+", version)
        or re.match(r"\d+\.\d+", version)
        or re.match(r"\d+\.\d+\.\d+", version)
    ) is not None

    assert is_valid_version

    def _(fn: CallableT) -> CallableT:

        versions_match = True
        for i, v in enumerate(version.split(".")):
            if sys.version_info[i] != int(v):
                versions_match = False
                break

        fn = pytest.mark.skipif(
            versions_match,
            reason=reason_str,
        )(fn)

        return fn

    return _


def run_command(cmd: str) -> str:
    output = subprocess.run(
        cmd,
        check=False,
        capture_output=True,
        text=True,
        shell=True,
    )
    if output.returncode != 0:
        print(
            "Failed to run command!",
            "\nstdout:",
            output.stdout,
            "\nstderr:",
            output.stderr,
            file=sys.stderr,
            sep="\n",
        )
        raise RuntimeError(f"Failed to run command: {redact_api_key(cmd)}")
    return output.stdout


def redact_api_key(cmd: str) -> str:
    # Redact the value of the `--api-key` CLI argument, replace it with `***`
    # Create a regex that replaces the argument following `--api-key`
    # with `***` (e.g. `--api-key my-api-key` -> `--api-key ***`)

    return re.sub(r"(--api-key\s+)(\S+)", r"\1***", cmd)


def deploy_to_connect(app_name: str, app_dir: str) -> str:
    if not api_key:
        raise RuntimeError("No api key found. Cannot deploy.")

    # check if connect app is already deployed to avoid duplicates
    connect_server_lookup_command = f"rsconnect content search --server {server_url} --api-key {api_key} --title-contains {app_name}"
    app_details = run_command(connect_server_lookup_command)
    connect_server_deploy = f"rsconnect deploy shiny {app_dir} --server {server_url} --api-key {api_key} --title {app_name} --verbose"
    # only if the app exists do we replace existing app with new version
    if json.loads(app_details):
        app_id = json.loads(app_details)[0]["guid"]
        connect_server_deploy += f" --app-id {app_id}"

    # Deploy to connect server
    run_command(connect_server_deploy)

    # look up content url in connect server once app is deployed
    output = run_command(connect_server_lookup_command)
    url = json.loads(output)[0]["content_url"]
    app_id = json.loads(output)[0]["guid"]

    # change visibility of app to public
    connect_app_url = f"{server_url}/__api__/v1/content/{app_id}"
    payload = '{"access_type":"all"}'
    headers = {
        "Authorization": f"Key {api_key}",
        "Accept": "application/json",
    }
    response = requests.request("PATCH", connect_app_url, headers=headers, data=payload)
    if response.status_code != 200:
        warnings.warn(
            f"Failed to change visibility of app. {response.text}",
            RuntimeWarning,
            stacklevel=1,
        )
        pytest.skip(
            "Skipping test as deployed app is not visible to public. Test is kept as it does confirm the app deployment has succeeded."
        )
        return

    return url


# TODO-future: Supress web browser from opening after deploying - https://github.com/rstudio/rsconnect-python/issues/462
def deploy_to_shinyapps(app_name: str, app_dir: str) -> str:
    # Deploy to shinyapps.io
    shinyapps_deploy = f"rsconnect deploy shiny {app_dir} --account {shinyappsio_name} --token {shinyappsio_token} --secret {shinyappsio_secret} --title {app_name} --verbose"
    run_command(shinyapps_deploy)
    return f"https://{shinyappsio_name}.shinyapps.io/{app_name}/"


# Since connect parses python packages, we need to get latest version of shiny on HEAD
def write_github_requirements_txt(app_dir: str) -> None:
    print("Writing github requirements.txt")
    app_requirements_file_path = os.path.join(app_dir, "app_requirements.txt")
    requirements_file_path = os.path.join(app_dir, "requirements.txt")
    git_cmd = subprocess.run(["git", "rev-parse", "HEAD"], stdout=subprocess.PIPE)
    git_hash = git_cmd.stdout.decode("utf-8").strip()
    with open(app_requirements_file_path) as f:
        requirements = f.read()
    with open(requirements_file_path, "w") as f:
        f.write(f"{requirements}\n")
        f.write(f"git+https://github.com/posit-dev/py-shiny.git@{git_hash}\n")


def write_pypi_requirements_txt(app_dir: str) -> None:
    print("Writing pypi requirements.txt")
    app_requirements_file_path = os.path.join(app_dir, "app_requirements.txt")
    requirements_file_path = os.path.join(app_dir, "requirements.txt")

    with open(app_requirements_file_path) as f:
        requirements = f.read()
    with open(requirements_file_path, "w") as f:
        f.write(f"{requirements}\n")
        f.write("shiny\n")


def assert_rsconnect_file_updated(file_path: str, min_mtime: float) -> None:
    """
    Asserts that the specified file has been updated since `min_mtime` (seconds since epoch).
    """
    mtime = os.path.getmtime(file_path)
    assert (
        mtime > min_mtime
    ), f"File '{file_path}' was not updated during app deployment which means the deployment failed"


def deploy_app(
    app_file_path: str,
    location: str,
    app_name: str,
) -> str:
    should_deploy_apps = os.environ.get("DEPLOY_APPS", "False") == "true"

    if not should_deploy_apps:
        pytest.skip("`DEPLOY_APPS` does not equal `true`")

    if not (run_on_ci and repo == "posit-dev/py-shiny"):
        pytest.skip("Not on CI and within posit-dev/py-shiny repo")

    app_dir = os.path.dirname(app_file_path)
    app_dir_name = os.path.basename(app_dir)

    # Use temporary directory to avoid modifying the original app directory
    # This allows us to run tests in parallel when deploying apps both modify the same rsconnect config file
    with tempfile.TemporaryDirectory("deploy_app") as tmpdir:

        # Creating a dir with same name instead of tmp to avoid issues
        # when deploying app to shinyapps.io using rsconnect package
        # since the rsconnect/*.json file needs the app_dir name to be same
        tmp_app_dir = os.path.join(tmpdir, app_dir_name)
        os.mkdir(tmp_app_dir)
        shutil.copytree(app_dir, tmp_app_dir, dirs_exist_ok=True)
        if should_use_github_requirements_txt:
            write_github_requirements_txt(tmp_app_dir)
        else:
            write_pypi_requirements_txt(tmp_app_dir)

        deployment_function = {
            "connect": deploy_to_connect,
            "shinyapps": deploy_to_shinyapps,
        }[location]

        pre_deployment_time = time.time()
        url = deployment_function(app_name, tmp_app_dir)
        tmp_rsconnect_config = os.path.join(
            tmp_app_dir, "rsconnect-python", f"{os.path.basename(tmp_app_dir)}.json"
        )
        assert_rsconnect_file_updated(tmp_rsconnect_config, pre_deployment_time)

        return url


def local_deploys_app_url_fixture(
    app_name: str,
    scope: ScopeName = "module",
):
    @pytest.fixture(scope=scope, params=[*deploy_locations, LOCAL_LOCATION])
    def fix_fn(request: pytest.FixtureRequest):

        app_file = os.path.join(os.path.dirname(request.path), "app.py")
        deploy_location = request.param

        if deploy_location == LOCAL_LOCATION:
            shinyapp_proc_gen = shiny_app_gen(app_file)
            # Return the `url`
            yield next(shinyapp_proc_gen).url
        elif deploy_location in deploy_locations:

            if deploy_location == "connect" and not (server_url and api_key):
                pytest.skip("Connect server url or api key not found. Cannot deploy.")
            if (
                deploy_location == "shinyapps"
                and shinyappsio_name
                and shinyappsio_token
                and shinyappsio_secret
            ):
                pytest.skip(
                    "Shinyapps.io name, token or secret not found. Cannot deploy."
                )

            app_url = deploy_app(
                app_file,
                deploy_location,
                app_name,
            )
            yield app_url

        else:
            raise ValueError(
                "Deploy location not a known location: '", deploy_location, "'"
            )

    return fix_fn



import os
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

from shiny import App, render, ui
from shiny.ui._theme_brand import bootstrap_colors

# TODO: Move this into the test that runs this app
os.environ["SHINY_BRAND_YML_RAISE_UNMAPPED"] = "true"
theme = ui.Theme.from_brand(__file__)
# theme = ui.Theme()
theme.add_rules((Path(__file__).parent / "_colors.scss").read_text())

app_ui = ui.page_navbar(
    ui.nav_panel(
        "Input Output Demo",
        ui.layout_sidebar(
            ui.sidebar(
                ui.input_slider("slider1", "Numeric Slider Input", 0, 11, 11),
                ui.input_numeric("numeric1", "Numeric Input Widget", 30),
                ui.input_date("date1", "Date Input Component", value="2024-01-01"),
                ui.input_switch("switch1", "Binary Switch Input", True),
                ui.input_radio_buttons(
                    "radio1",
                    "Radio Button Group",
                    choices=["Option A", "Option B", "Option C", "Option D"],
                ),
                ui.input_action_button("action1", "Action Button"),
            ),
            ui.layout_columns(
                ui.value_box(
                    "Metric 1",
                    "100",
                    theme="primary",
                ),
                ui.value_box(
                    "Metric 2",
                    "200",
                    theme="secondary",
                ),
                ui.value_box(
                    "Metric 3",
                    "300",
                    theme="info",
                ),
            ),
            ui.card(
                ui.card_header("Plot Output"),
                ui.output_plot("plot1"),
            ),
            ui.card(
                ui.card_header("Text Output"),
                ui.output_text_verbatim("out_text1"),
            ),
        ),
    ),
    ui.nav_panel(
        "Widget Gallery",
        ui.layout_column_wrap(
            ui.card(
                ui.card_header("Button Variants"),
                ui.input_action_button("btn1", "Default"),
                ui.input_action_button("btn2", "Primary", class_="btn-primary"),
                ui.input_action_button("btn3", "Secondary", class_="btn-secondary"),
                ui.input_action_button("btn4", "Info", class_="btn-info"),
                ui.input_action_button("btn5", "Success", class_="btn-success"),
                ui.input_action_button("btn6", "Warning", class_="btn-warning"),
                ui.input_action_button("btn7", "Danger", class_="btn-danger"),
            ),
            ui.card(
                ui.card_header("Radio Button Examples"),
                ui.input_radio_buttons(
                    "radio2",
                    "Standard Radio Group",
                    ["Selection 1", "Selection 2", "Selection 3"],
                ),
                ui.input_radio_buttons(
                    "radio3",
                    "Inline Radio Group",
                    ["Option 1", "Option 2", "Option 3"],
                    inline=True,
                ),
            ),
            ui.card(
                ui.card_header("Checkbox Examples"),
                ui.input_checkbox_group(
                    "check1",
                    "Standard Checkbox Group",
                    ["Item 1", "Item 2", "Item 3"],
                ),
                ui.input_checkbox_group(
                    "check2",
                    "Inline Checkbox Group",
                    ["Choice A", "Choice B", "Choice C"],
                    inline=True,
                ),
            ),
            ui.card(
                ui.card_header("Select Input Widgets"),
                ui.input_selectize(
                    "select1",
                    "Selectize Input",
                    ["Selection A", "Selection B", "Selection C"],
                ),
                ui.input_select(
                    "select2",
                    "Multiple Select Input",
                    ["Item X", "Item Y", "Item Z"],
                    multiple=True,
                ),
            ),
            ui.card(
                ui.card_header("Text Input Widgets"),
                ui.input_text("text1", "Text Input"),
                ui.input_text_area(
                    "textarea1",
                    "Text Area Input",
                    "Default text content for the text area widget",
                ),
                ui.input_password("password1", "Password Input"),
            ),
            width=300,
            heights_equal=False,
        ),
    ),
    ui.nav_panel(
        "Colors",
        ui.fill.as_fill_item(
            ui.div(
                ui.div(ui.output_ui("ui_colors"), class_="container-sm"),
                class_="overflow-y-auto",
            )
        ),
    ),
    ui.nav_panel(
        "Documentation",
        ui.fill.as_fill_item(
            ui.div(
                ui.div(
                    ui.markdown(
                        """
                _Just in case it isn't obvious, this text was written by an LLM._

                # Component Documentation

                The Shiny for Python framework, available at [shiny.posit.co/py](https://shiny.posit.co/py/),
                provides a comprehensive set of UI components for building interactive web applications. These
                components are designed with **consistency and usability** in mind, making it easier for
                developers to create professional-grade applications.

                Our framework implements the `ui.page_navbar()` component as the primary navigation structure,
                allowing for intuitive organization of content across multiple views. Each view can contain
                various input and output elements, managed through the `ui.card()` container system.

                ## Component Architecture

                *The architecture of our application framework* emphasizes modularity and reusability. Key
                components like `ui.value_box()` and `ui.layout_column_wrap()` work together to create
                structured, responsive layouts that adapt to different screen sizes.

                <table class="table table-striped">
                <thead>
                <tr>
                <th>Component</th>
                <th>Implementation</th>
                <th>Use Case</th>
                <th>Status</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>Value Box</td>
                <td><code>ui.value_box()</code></td>
                <td>Metric Display</td>
                <td>Production Ready</td>
                </tr>
                <tr>
                <td>Card</td>
                <td><code>ui.card()</code></td>
                <td>Content Container</td>
                <td>Production Ready</td>
                </tr>
                <tr>
                <td>Layout</td>
                <td><code>ui.layout_column_wrap()</code></td>
                <td>Component Organization</td>
                <td>Production Ready</td>
                </tr>
                <tr>
                <td>Navigation</td>
                <td><code>ui.page_navbar()</code></td>
                <td>Page Structure</td>
                <td>Production Ready</td>
                </tr>
                </tbody>
                </table>

                ## Implementation Best Practices

                When implementing components, maintain consistent patterns in your code. Use the
                `@render` decorators for output functions and follow the reactive programming model
                with `@reactive.effect` for side effects.

                Error handling should be implemented at both the UI and server levels. For input
                validation, use the `req()` function to ensure all required values are present
                before processing.

                ## Corporate Brand Guidelines

                Effective corporate brand guidelines should accomplish several key objectives:

                1. **Visual Consistency**: Establish a clear color palette using our theming system.
                Primary colors should be defined using `class_="btn-primary"` and similar Bootstrap
                classes.

                2. *Typography Standards*: Maintain consistent font usage across all text elements.
                Headers should use the built-in styling provided by the `ui.card_header()` component.

                3. `Component Styling`: Apply consistent styling to UI elements such as buttons,
                cards, and value boxes. Use the theme parameter in components like
                `ui.value_box(theme="primary")`.

                4. **Layout Principles**: Follow a grid-based layout system using
                `ui.layout_column_wrap()` with appropriate width parameters to ensure consistent
                spacing and alignment.

                5. *Responsive Design*: Implement layouts that adapt gracefully to different screen
                sizes using the `fillable` parameter in page components.

                Remember that brand guidelines should serve as a framework for consistency while
                remaining flexible enough to accommodate future updates and modifications to the
                application interface.
                """
                    ),
                    class_="container-sm ",
                ),
                class_="overflow-y-auto",
            )
        ),
    ),
    ui.nav_spacer(),
    ui.nav_control(ui.input_dark_mode(id="color_mode")),
    title="brand.yml Demo",
    fillable=True,
    theme=theme,
)


def server(input, output, session):
    @render.plot
    def plot1():
        colors = {
            "foreground": theme.brand.color.foreground,
            "background": theme.brand.color.background,
            "primary": theme.brand.color.primary,
        }

        if theme.brand.color:
            colors.update(theme.brand.color.to_dict("theme"))

        if input.color_mode() == "dark":
            bg = colors["foreground"]
            fg = colors["background"]
            colors.update({"foreground": fg, "background": bg})

        x = np.linspace(0, input.numeric1(), 100)
        y = np.sin(x) * input.slider1()
        fig, ax = plt.subplots(facecolor=colors["background"])
        ax.plot(x, y, color=colors["primary"])
        ax.set_title("Sine Wave Output", color=colors["foreground"])
        ax.set_facecolor(colors["background"])
        ax.tick_params(colors=colors["foreground"])
        for spine in ax.spines.values():
            spine.set_edgecolor(colors["foreground"])
            spine.set_alpha(0.25)
        return fig

    @render.text
    def out_text1():
        return "\n".join(
            ["def example_function():", '    return "Function output text"']
        )

    @render.ui
    def ui_colors():
        colors = []
        # Replicates: https://getbootstrap.com/docs/5.3/customize/color/#all-colors
        # Source: https://github.com/twbs/bootstrap/blob/6e1f75f4/site/content/docs/5.3/customize/color.md?plain=1#L395-L409
        for color in ["gray", *bootstrap_colors]:
            if color in ["white", "black"]:
                continue

            colors += [
                ui.div(
                    ui.div(color, class_=f"p-3 mb-2 position-relative bd-{color}-500"),
                    *[
                        ui.div(f"{color}-{r}", class_=f"p-3 bd-{color}-{r}")
                        for r in range(100, 1000, 100)
                    ],
                    class_="mb-3",
                )
            ]

        return ui.TagList(
            ui.div(
                *[
                    ui.div(
                        ui.div(
                            color, class_=f"p-3 mb-2 position-relative text-bg-{color}"
                        ),
                        class_="col-md-3 mb-3",
                    )
                    for color in [
                        "primary",
                        "secondary",
                        "dark",
                        "light",
                        "info",
                        "success",
                        "warning",
                        "danger",
                    ]
                ],
                class_="row font-monospace",
            ),
            ui.div(
                *[
                    ui.div(
                        ui.div(color, class_=f"p-3 mb-2 position-relative bd-{color}"),
                        class_="col-md-3 mb-3",
                    )
                    for color in ["black", "white", "foreground", "background"]
                ],
                class_="row font-monospace",
            ),
            ui.layout_column_wrap(*colors, class_="font-monospace"),
        )


app = App(app_ui, server)



from shiny.express import ui

ui.h2("A basic absolute panel example")

with ui.panel_absolute(draggable=True, width="300px", right="50px", top="25%"):
    with ui.panel_well():
        "Drag me around!"
        ui.input_slider("n", "N", min=0, max=100, value=20)



from pathlib import Path

from shiny import App, ui

js_file = Path(__file__).parent / "js" / "app.js"

app_ui = ui.page_fluid(
    "If you see this page before 'OK'-ing the alert box, something went wrong",
    ui.include_js(js_file),
)


app = App(app_ui, None)



from shiny.express import input, render, ui

ui.page_opts(title="Select Inputs Kitchensink")

fruits = ["Apple", "Banana", "Cherry", "Date", "Elderberry"]
fruits_dict = {
    "apple": "Apple",
    "banana": "Banana",
    "cherry": "Cherry",
    "date": "Date",
    "elderberry": "Elderberry",
}
fruits_grouped_dict = {
    "Citrus": {
        "Orange": "Sweet and tangy",
        "Lemon": "Zesty and refreshing",
        "Lime": "Bright and tart",
    },
    "Berries": {
        "Strawberry": "Juicy and sweet",
        "Blueberry": "Tiny and antioxidant-rich",
        "Raspberry": "Delicate and slightly tart",
    },
}


with ui.card():
    ui.input_select("basic_select", "Default select", fruits)

    @render.code
    def basic_result_txt():
        return input.basic_select()


with ui.card():
    ui.input_select("multi_select", "Multiple Select", fruits, multiple=True)

    @render.code
    def multi_result_txt():
        return ", ".join(input.multi_select())


with ui.card():
    ui.input_select(
        "select_with_selected", "Select with selected", fruits, selected="Cherry"
    )

    @render.code
    def select_with_selected_txt():
        return str(input.select_with_selected())


with ui.card():
    ui.input_select("width_select", "Select with Custom Width", fruits, width="400px")

    @render.code
    def width_result_txt():
        return str(input.width_select())


with ui.card():
    ui.input_select(
        "select_with_labels",
        "Select with labels",
        fruits_dict,
    )

    @render.code
    def select_with_labels_txt():
        return str(input.select_with_labels())


with ui.card():
    ui.input_select(
        "select_with_custom_size_and_dict",
        "Select with custom size and dict",
        fruits_grouped_dict,
        size="4",
    )

    @render.code
    def select_with_custom_size_and_dict_txt():
        return str(input.select_with_custom_size_and_dict())



from shared import mtcars

from shiny import App, reactive, render, ui

app_ui = ui.page_fillable(
    ui.card(
        ui.layout_column_wrap(
            ui.input_action_button("btn", "Filter on columns 0, 1, and 3"),
            ui.input_action_button("reset", "Reset column filters"),
            fill=False,
        ),
        ui.output_data_frame("df"),
    ),
)


def server(input, output, session):
    data = reactive.value(mtcars.iloc[:, range(4)])

    @render.data_frame
    def df():
        return render.DataGrid(data(), filters=True)

    @reactive.effect
    @reactive.event(input.reset)
    async def _():
        await df.update_filter(None)

    @reactive.effect
    @reactive.event(input.btn)
    async def _():
        await df.update_filter(
            [
                {"col": 0, "value": [19, 25]},
                {"col": 1, "value": [None, 6]},
                {"col": 3, "value": [100, None]},
            ]
        )


app = App(app_ui, server, debug=True)



# See https://github.com/microsoft/playwright-python/issues/1532
# pyright: reportUnknownMemberType=false

from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_async_app(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    controller.InputTextArea(page, "value").set("Hello\nGoodbye")
    controller.InputActionButton(page, "go").click()

    # TODO-future; Make into proper class
    progress = page.locator("#shiny-notification-panel")
    expect(progress).to_be_visible()
    expect(progress).to_contain_text("Calculating...")

    controller.OutputTextVerbatim(page, "hash_output").expect_value(
        "2e220fb9d401bf832115305b9ae0277e7b8b1a9368c6526e450acd255e0ec0c2", timeout=2000
    )

    expect(progress).to_be_hidden()



import pandas  # noqa: F401 (this line needed for Shinylive to load plotly.express)
import plotly.express as px
from shinywidgets import render_widget

from shiny import reactive, req
from shiny.express import render, ui

# Load the Gapminder dataset
df = px.data.gapminder()

# Prepare a summary DataFrame
summary_df = (
    df.groupby("country")
    .agg(
        {
            "pop": ["min", "max", "mean"],
            "lifeExp": ["min", "max", "mean"],
            "gdpPercap": ["min", "max", "mean"],
        }
    )
    .reset_index()
)

summary_df.columns = ["_".join(col).strip() for col in summary_df.columns.values]
summary_df.rename(columns={"country_": "country"}, inplace=True)

# Set up the UI

ui.page_opts(fillable=True)

ui.markdown(
    "**Instructions**: Select one or more countries in the table below to see more information."
)

with ui.layout_columns(col_widths=[12, 6, 6]):
    with ui.card(height="400px"):

        @render.data_frame
        def summary_data():
            return render.DataGrid(summary_df.round(2), selection_mode="rows")

    with ui.card(height="400px"):

        @render_widget
        def country_detail_pop():
            return px.line(
                filtered_df(),
                x="year",
                y="pop",
                color="country",
                title="Population Over Time",
            )

    with ui.card(height="400px"):

        @render_widget
        def country_detail_percap():
            return px.line(
                filtered_df(),
                x="year",
                y="gdpPercap",
                color="country",
                title="GDP per Capita Over Time",
            )


@reactive.calc
def filtered_df():
    data_selected = summary_data.data_view(selected=True)
    req(not data_selected.empty)
    countries = data_selected["country"]

    # Filter data for selected countries
    return df[df["country"].isin(countries)]



from palmerpenguins import load_penguins_raw  # pyright: ignore[reportMissingTypeStubs]

from shiny import App, Inputs, Outputs, Session, render, ui

df = load_penguins_raw()

df["Species"] = df["Species"].apply(lambda x: ui.HTML(f"<u>{x}</u>"))  # pyright: ignore


app_ui = ui.page_fluid(
    ui.h2("Palmer Penguins"),
    ui.output_data_frame("penguins_df"),
)


def server(input: Inputs, output: Outputs, session: Session) -> None:
    @render.data_frame
    def penguins_df():
        return render.DataGrid(
            data=df,  # pyright: ignore[reportUnknownArgumentType]
            editable=True,
            selection_mode="rows",
        )


app = App(app_ui, server)



from shiny import App, render, ui

app_ui = ui.page_fluid(
    ui.panel_title("Hello Shiny!"),
    ui.input_slider("n", "N", 0, 100, 20),
    ui.output_text_verbatim("txt"),
)


def server(input, output, session):
    @render.text
    def txt():
        return f"n*2 is {input.n() * 2}"


app = App(app_ui, server)



import time

from palmerpenguins import load_penguins_raw  # pyright: ignore[reportMissingTypeStubs]

from shiny import App, Inputs, Outputs, Session, render, ui
from shiny.render import CellPatch

app_ui = ui.page_fluid(
    ui.h2("Palmer Penguins"),
    ui.output_data_frame("penguins_df"),
)


def server(input: Inputs, output: Outputs, session: Session) -> None:
    @render.data_frame
    def penguins_df():
        return render.DataGrid(
            data=load_penguins_raw(),  # pyright: ignore[reportUnknownArgumentType]
            filters=True,
            editable=True,
        )

    @penguins_df.set_patch_fn
    async def upgrade_patch(
        *,
        patch: CellPatch,
    ):
        # Slow down change so that we can test for "editing" state
        time.sleep(2)
        return patch["value"]


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.tags.p("The first checkbox group controls the second"),
    ui.input_checkbox_group(
        "inCheckboxGroup", "Input checkbox", ["Item A", "Item B", "Item C"]
    ),
    ui.input_checkbox_group(
        "inCheckboxGroup2", "Input checkbox 2", ["Item A", "Item B", "Item C"]
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    def _():
        x = input.inCheckboxGroup()

        # Can also set the label and select items
        ui.update_checkbox_group(
            "inCheckboxGroup2",
            label="Checkboxgroup label " + str(len(x)),
            choices=x,
            selected=x,
        )


app = App(app_ui, server)



from __future__ import annotations

from pathlib import Path

__all__ = (
    "page_sidebar",
    "page_navbar",
    "page_fillable",
    "page_fluid",
    "page_fixed",
    "page_bootstrap",
    "page_auto",
    "page_output",
)

from copy import copy
from typing import Any, Callable, Literal, Optional, Sequence, cast

from htmltools import (
    MetadataNode,
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    TagList,
    css,
    div,
    head_content,
    tags,
)

from .._docstring import add_example, no_example
from .._namespaces import resolve_id_or_none
from ..types import MISSING, MISSING_TYPE, NavSetArg
from ._bootstrap import panel_title
from ._html_deps_external import Theme, ThemeProvider, shiny_page_theme_deps
from ._html_deps_py_shiny import page_output_dependency
from ._html_deps_shinyverse import components_dependencies
from ._navs import NavMenu, NavPanel, navset_bar
from ._sidebar import Sidebar, SidebarOpen, layout_sidebar
from ._tag import consolidate_attrs
from ._utils import get_window_title
from .css import CssUnit, as_css_padding, as_css_unit
from .fill._fill import as_fill_item, as_fillable_container

page_sidebar_default: SidebarOpen = SidebarOpen(desktop="open", mobile="always")


@add_example()
def page_sidebar(
    sidebar: Sidebar,
    *args: TagChild | TagAttrs,
    title: Optional[str | Tag | TagList] = None,
    fillable: bool = False,
    fillable_mobile: bool = False,
    window_title: str | MISSING_TYPE = MISSING,
    lang: Optional[str] = None,
    theme: Optional[str | Path | Theme | ThemeProvider] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a page with a sidebar and a title.

    Parameters
    ----------
    sidebar
        Content to display in the sidebar.
    *args
        UI elements.
    title
        A title to display at the top of the page.
    fillable
        Whether or not the main content area should be considered a fillable
        (i.e., flexbox) container.
    fillable_mobile
        Whether or not ``fillable`` should apply on mobile devices.
    window_title
        The browser's window title (defaults to the host URL of the page). Can also be
        set as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    theme
        A custom Shiny theme created using the :class:`~shiny.ui.Theme` class, or a path
        to a local or online CSS file that will replace the Bootstrap CSS bundled by
        default with a Shiny app. This file should be a complete `bootstrap.css` or
        `bootstrap.min.css` file.

        For advanced uses, you can also pass a :class:`~htmltools.Tagifiable` object.
        In this case, Shiny will suppress the default Bootstrap CSS.

        To modify the theme of an app without replacing the Bootstrap CSS entirely, use
        :func:`~shiny.ui.include_css` to add custom CSS.
    **kwargs
        Additional attributes passed to :func:`~shiny.ui.layout_sidebar`.

    Returns
    -------
    :
        A UI element.
    """

    if isinstance(title, str):
        title = tags.h1(title, class_="bslib-page-title navbar-brand")

    if title is not None:
        navbar_title = tags.div(
            tags.div(title, class_="container-fluid"),
            class_="navbar navbar-static-top",
        )
    else:
        navbar_title = None

    if not isinstance(sidebar, Sidebar):
        raise TypeError(
            "`sidebar=` is not a `Sidebar` instance. Use `ui.sidebar(...)` to create one."
        )

    if sidebar._default_open != page_sidebar_default:
        sidebar = copy(sidebar)
        sidebar._default_open = page_sidebar_default

    attrs, children = consolidate_attrs(*args, **kwargs)

    return page_fillable(
        {"class": "bslib-page-sidebar"},
        navbar_title,
        layout_sidebar(
            sidebar,
            page_main_container(*children),
            attrs,
            fillable=fillable,
            border=False,
            border_radius=False,
        ),
        get_window_title(title, window_title=window_title),
        padding=0,
        gap=0,
        lang=lang,
        theme=theme,
        fillable_mobile=fillable_mobile,
    )


def page_main_container(*args: TagChild) -> Tag:
    main = tags.main({"class": "bslib-page-main bslib-gap-spacing"}, *args)
    return as_fillable_container(as_fill_item(main))


@no_example()
def page_navbar(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    title: Optional[str | Tag | TagList] = None,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    sidebar: Optional[Sidebar] = None,
    # Only page_navbar gets enhanced treatement for `fillable`
    # If an `*args`'s `data-value` attr string is in `fillable`, then the component is fillable
    fillable: bool | list[str] = False,
    fillable_mobile: bool = False,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    position: Literal["static-top", "fixed-top", "fixed-bottom"] = "static-top",
    header: Optional[TagChild] = None,
    footer: Optional[TagChild] = None,
    bg: Optional[str] = None,
    inverse: bool = False,
    underline: bool = True,
    collapsible: bool = True,
    fluid: bool = True,
    window_title: str | MISSING_TYPE = MISSING,
    lang: Optional[str] = None,
    theme: Optional[str | Path | Theme | ThemeProvider] = None,
) -> Tag:
    """
    Create a page with a navbar and a title.

    Parameters
    ----------
    *args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match its
        ``value``).
    sidebar
        A :func:`~shiny.ui.sidebar` component to display on every page.
    fillable
        Whether or not the main content area should be considered a fillable
        (i.e., flexbox) container.
    fillable_mobile
        Whether or not ``fillable`` should apply on mobile devices.
    gap
        A CSS length unit defining the gap (i.e., spacing) between elements provided to
        `*args`. This value is only used when the navbar is _fillable_.
    padding
        Padding to use for the body. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. The length
        can be between one and four. If one, then that value will be used for all four
        sides. If two, then the first value will be used for the top and bottom, while
        the second value will be used for left and right. If three, then the first will
        be used for top, the second will be left and right, and the third will be
        bottom. If four, then the values will be interpreted as top, right, bottom, and
        left respectively. This value is only used when the navbar is _fillable_.
    position
        Determines whether the navbar should be displayed at the top of the page with
        normal scrolling behavior ("static-top"), pinned at the top ("fixed-top"), or
        pinned at the bottom ("fixed-bottom"). Note that using "fixed-top" or
        "fixed-bottom" will cause the navbar to overlay your body content, unless you
        add padding (e.g., ``tags.style("body {padding-top: 70px;}")``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    bg
        Background color of the navbar (a CSS color).
    inverse
        Either ``True`` for a light text color or ``False`` for a dark text color.
    collapsible
        ``True`` to automatically collapse the elements into an expandable menu on mobile devices or narrow window widths.
    fluid
        ``True`` to use fluid layout; ``False`` to use fixed layout.
    window_title
        The browser's window title (defaults to the host URL of the page). Can also be
        set as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    theme
        A custom Shiny theme created using the :class:`~shiny.ui.Theme` class, or a path
        to a local or online CSS file that will replace the Bootstrap CSS bundled by
        default with a Shiny app. This file should be a complete `bootstrap.css` or
        `bootstrap.min.css` file.

        For advanced uses, you can also pass a :class:`~htmltools.Tagifiable` object.
        In this case, Shiny will suppress the default Bootstrap CSS.

        To modify the theme of an app without replacing the Bootstrap CSS entirely, use
        :func:`~shiny.ui.include_css` to add custom CSS.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.nav`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.page_fluid`

    Example
    -------
    See :func:`~shiny.ui.nav`.
    """
    pageClass = "bslib-page-navbar"

    if sidebar is not None:
        if not isinstance(sidebar, Sidebar):
            raise TypeError(
                "`sidebar=` is not a `Sidebar` instance. Use `ui.sidebar(...)` to create one."
            )

        pageClass += " has-page-sidebar"
        if sidebar._default_open != page_sidebar_default:
            sidebar = copy(sidebar)
            sidebar._default_open = page_sidebar_default

    tagAttrs: TagAttrs = {"class": pageClass}

    navbar = navset_bar(
        *args,
        title=title,
        id=resolve_id_or_none(id),
        selected=selected,
        sidebar=sidebar,
        fillable=fillable,
        gap=gap,
        padding=padding,
        position=position,
        header=header,
        footer=footer,
        bg=bg,
        inverse=inverse,
        underline=underline,
        collapsible=collapsible,
        fluid=fluid,
    )
    # This is a page-level navbar, so opt into page-level layouts (in particular for
    # navbar with a global sidebar)
    navbar._is_page_level = True

    page_args = (
        tagAttrs,
        navbar,
        get_window_title(title, window_title=window_title),
    )
    page_kwargs = {
        "title": None,
        "lang": lang,
    }

    # If a sidebar is provided, we want the layout_sidebar(fill = TRUE) component
    # (which is a sibling of the <nav>) to always fill the page
    if fillable is False and sidebar is None:
        return page_bootstrap(
            *page_args,
            theme=theme,
            **page_kwargs,
        )

    else:
        return page_fillable(
            *page_args,
            fillable_mobile=fillable_mobile,
            padding=0,
            gap=0,
            theme=theme,
            **page_kwargs,
        )


@no_example()
def page_fillable(
    *args: TagChild | TagAttrs,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    gap: Optional[CssUnit] = None,
    fillable_mobile: bool = False,
    title: Optional[str] = None,
    lang: Optional[str] = None,
    theme: Optional[str | Path | Theme | ThemeProvider] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a fillable page.

    Parameters
    ----------
    *args
        UI elements.
    padding
        Padding to use for the body. See :func:`~shiny.ui.css.as_css_padding`
        for more details.
    fillable_mobile
        Whether or not the page should fill the viewport's height on mobile devices
        (i.e., narrow windows).
    gap
        A CSS length unit passed through :func:`~shiny.ui.css.as_css_unit`
        defining the `gap` (i.e., spacing) between elements provided to `*args`.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    theme
        A custom Shiny theme created using the :class:`~shiny.ui.Theme` class, or a path
        to a local or online CSS file that will replace the Bootstrap CSS bundled by
        default with a Shiny app. This file should be a complete `bootstrap.css` or
        `bootstrap.min.css` file.

        For advanced uses, you can also pass a :class:`~htmltools.Tagifiable` object.
        In this case, Shiny will suppress the default Bootstrap CSS.

        To modify the theme of an app without replacing the Bootstrap CSS entirely, use
        :func:`~shiny.ui.include_css` to add custom CSS.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.page_fluid`
    * :func:`~shiny.ui.page_fixed`
    """
    attrs, children = consolidate_attrs(*args, **kwargs)

    style = css(padding=as_css_padding(padding), gap=as_css_unit(gap))

    page = page_bootstrap(
        head_content(tags.style("html { height: 100%; }")),
        {"class": "bslib-page-fill bslib-gap-spacing", "style": style},
        {"class": "bslib-flow-mobile"} if not fillable_mobile else None,
        attrs,
        *children,
        components_dependencies(),
        title=title,
        lang=lang,
        theme=theme,
    )

    # page returns a <html> tag, but we need to make the <body> fillable
    body = page.children[1]
    if not isinstance(body, Tag) or body.name != "body":
        raise ValueError("Expected a <body> tag")

    page.children[1] = as_fillable_container(body)

    return page


@add_example()
def page_fluid(
    *args: TagChild | TagAttrs,
    title: Optional[str] = None,
    lang: Optional[str] = None,
    theme: Optional[str | Path | Theme | ThemeProvider] = None,
    **kwargs: str,
) -> Tag:
    """
    Create a fluid page.

    Parameters
    ----------
    *args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    theme
        A custom Shiny theme created using the :class:`~shiny.ui.Theme` class, or a path
        to a local or online CSS file that will replace the Bootstrap CSS bundled by
        default with a Shiny app. This file should be a complete `bootstrap.css` or
        `bootstrap.min.css` file.

        For advanced uses, you can also pass a :class:`~htmltools.Tagifiable` object.
        In this case, Shiny will suppress the default Bootstrap CSS.

        To modify the theme of an app without replacing the Bootstrap CSS entirely, use
        :func:`~shiny.ui.include_css` to add custom CSS.
    **kwargs
        Attributes on the page level container.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.page_fixed`
    * :func:`~shiny.ui.page_bootstrap`
    * :func:`~shiny.ui.page_navbar`
    """

    return page_bootstrap(
        div({"class": "container-fluid"}, *args, **kwargs),
        title=title,
        lang=lang,
        theme=theme,
    )


@add_example()
def page_fixed(
    *args: TagChild | TagAttrs,
    title: Optional[str] = None,
    lang: Optional[str] = None,
    theme: Optional[str | Path | Theme | ThemeProvider] = None,
    **kwargs: str,
) -> Tag:
    """
    Create a fixed page.

    Parameters
    ----------
    *args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    theme
        A custom Shiny theme created using the :class:`~shiny.ui.Theme` class, or a path
        to a local or online CSS file that will replace the Bootstrap CSS bundled by
        default with a Shiny app. This file should be a complete `bootstrap.css` or
        `bootstrap.min.css` file.

        For advanced uses, you can also pass a :class:`~htmltools.Tagifiable` object.
        In this case, Shiny will suppress the default Bootstrap CSS.

        To modify the theme of an app without replacing the Bootstrap CSS entirely, use
        :func:`~shiny.ui.include_css` to add custom CSS.
    **kwargs
        Attributes on the page level container.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.page_fluid`
    * :func:`~shiny.ui.page_bootstrap`
    * :func:`~shiny.ui.page_navbar`
    """

    return page_bootstrap(
        div({"class": "container"}, *args, **kwargs),
        title=title,
        lang=lang,
        theme=theme,
    )


# TODO: implement theme (just Bootswatch for now?)
@no_example()
def page_bootstrap(
    *args: TagChild | TagAttrs,
    title: Optional[str] = None,
    lang: Optional[str] = None,
    theme: Optional[str | Path | Theme | ThemeProvider] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a Bootstrap UI page container.

    Parameters
    ----------
    *args
        UI elements.
    title
        The browser window title (defaults to the host URL of the page). Can also be set
        as a side effect via :func:`~shiny.ui.panel_title`.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    theme
        A custom Shiny theme created using the :class:`~shiny.ui.Theme` class, or a path
        to a local or online CSS file that will replace the Bootstrap CSS bundled by
        default with a Shiny app. This file should be a complete `bootstrap.css` or
        `bootstrap.min.css` file.

        For advanced uses, you can also pass a :class:`~htmltools.Tagifiable` object.
        In this case, Shiny will suppress the default Bootstrap CSS.

        To modify the theme of an app without replacing the Bootstrap CSS entirely, use
        :func:`~shiny.ui.include_css` to add custom CSS.
    **kwargs
        Attributes on the the `<body>` tag.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :func:`~shiny.ui.page_fluid`
    * :func:`~shiny.ui.page_navbar`
    """
    head = tags.title(title) if title else None
    return tags.html(
        tags.head(head),
        tags.body(shiny_page_theme_deps(theme), *args, **kwargs),
        lang=lang,
    )


@no_example()
def page_auto(
    *args: TagChild | TagAttrs,
    title: str | MISSING_TYPE = MISSING,
    window_title: str | MISSING_TYPE = MISSING,
    lang: str | MISSING_TYPE = MISSING,
    theme: str | Path | Theme | ThemeProvider | MISSING_TYPE = MISSING,
    fillable: bool | MISSING_TYPE = MISSING,
    full_width: bool = False,
    page_fn: Callable[..., Tag] | None = None,
    **kwargs: object,
) -> Tag:
    """
    A page container which automatically decides which page function to use.

    If there is a top-level :func:`~shiny.ui.nav_panel`, :func:`~shiny.ui.page_auto`
    will use :func:`~shiny.ui.page_navbar`. Otherwise, if there is a top-level sidebar,
    :func:`~shiny.ui.page_sidebar` is used.

    If there are neither top-level nav panels nor sidebars, this will use the `fillable`
    and `full_width` arguments to determine which page function to use:

    1. If `fillable` is `True`, :func:`~shiny.ui.page_fillable` is used.
    2. Otherwise, if `full_width` is `True`, :func:`~shiny.ui.page_fluid` is used.
    3. If neither are `True`, :func:`~shiny.ui.page_fixed` is used.

    Parameters
    ----------
    *args
        UI elements. These are used to determine which page function to use, and they
        are also passed along to that page function.
    title
        A title shown on the page.
    window_title
        The browser window title. If no value is provided, this will use the value of
        ``title``.
    lang
        ISO 639-1 language code for the HTML page, such as ``"en"`` or ``"ko"``. This
        will be used as the lang in the ``<html>`` tag, as in ``<html lang="en">``. The
        default, `None`, results in an empty string.
    theme
        A custom Shiny theme created using the :class:`~shiny.ui.Theme` class, or a path
        to a local or online CSS file that will replace the Bootstrap CSS bundled by
        default with a Shiny app. This file should be a complete `bootstrap.css` or
        `bootstrap.min.css` file.

        For advanced uses, you can also pass a :class:`~htmltools.Tagifiable` object.
        In this case, Shiny will suppress the default Bootstrap CSS.

        To modify the theme of an app without replacing the Bootstrap CSS entirely, use
        :func:`~shiny.ui.include_css` to add custom CSS.
    fillable
        If there is a top-level sidebar or nav, then the value is passed through to the
        :func:`~shiny.ui.page_sidebar` or :func:`~shiny.ui.page_navbar` function.
        Otherwise, if ``True``, use :func:`~shiny.ui.page_fillable`, where the content
        fills the window; if ``False`` (the default), the value of ``full_width`` will
        determine which page function is used.
    full_width
        This has an effect only if there are no sidebars or top-level navs, and
        ``fillable`` is ``False``. If this is ``False`` (the default), use use
        :func:`~shiny.ui.page_fixed`; if ``True``, use :func:`~shiny.ui.page_fillable`.
    page_fn
        The page function to use. If ``None`` (the default), will automatically choose
        one based on the arguments provided. If not ``None``, this will override all
        heuristics for choosing page functions.
    **kwargs
        Additional arguments, which are passed to the page function.

    Returns
    -------
    :
        A UI element.
    """
    if not isinstance(title, MISSING_TYPE):
        kwargs["title"] = title
    if not isinstance(window_title, MISSING_TYPE):
        kwargs["window_title"] = window_title
    if not isinstance(lang, MISSING_TYPE):
        kwargs["lang"] = lang
    if not isinstance(theme, MISSING_TYPE):
        kwargs["theme"] = theme

    # Presence of a top-level nav items and/or sidebar determines the page function
    navs = [x for x in args if isinstance(x, (NavPanel, NavMenu))]
    sidebars = [x for x in args if isinstance(x, Sidebar)]

    nNavs = len(navs)
    nSidebars = len(sidebars)
    if page_fn is None:
        if nNavs == 0:
            if nSidebars == 0:
                if isinstance(fillable, MISSING_TYPE):
                    fillable = False

                if fillable:
                    page_fn = _page_auto_fillable
                elif full_width:
                    page_fn = _page_auto_fluid
                else:
                    page_fn = _page_auto_fixed

            elif nSidebars == 1:
                if not isinstance(fillable, MISSING_TYPE):
                    kwargs["fillable"] = fillable

                # page_sidebar() needs sidebar to be the first arg
                # TODO: Change page_sidebar() to remove `sidebar` and accept a sidebar as a
                # *arg.
                page_fn = page_sidebar
                args = tuple(sidebars + [x for x in args if x not in sidebars])

            else:
                raise NotImplementedError(
                    "Multiple top-level sidebars not allowed. Did you meant to wrap each one in layout_sidebar()?"
                )

        # At least one nav
        else:
            if not isinstance(fillable, MISSING_TYPE):
                kwargs["fillable"] = fillable

            if nSidebars == 0:
                # TODO: what do we do when nArgs != nNavs? Just let page_navbar handle it (i.e. error)?
                page_fn = page_navbar

            elif nSidebars == 1:
                # TODO: change page_navbar() to remove `sidebar` and accept a sidebar as a
                # *arg.
                page_fn = page_navbar
                args = tuple([x for x in args if x not in sidebars])
                kwargs["sidebar"] = sidebars[0]

            else:
                raise NotImplementedError(
                    "Multiple top-level sidebars not allowed in combination with top-level navs."
                )

    # If we got here, _page_fn is not None, but the type checker needs a little help.
    page_fn = cast(Callable[..., Tag], page_fn)
    return page_fn(*args, **kwargs)


# For `page_fillable`, `page_fluid`, and `page_fixed`, the `title` arg sets the window
# title, but doesn't add anything visible on the page.
#
# In contrast, for `page_auto`, the `title` arg adds a title panel to the page, and the
# `window_title` arg sets the window title.
#
# The wrapper functions below provide the `page_auto` interface, where `title` to add a
# title panel to the page, and `window_title` to set the title of the window. If `title`
# is provided but `window_title` is not, then `window_title` is set to the value of
# `title`.
def _page_auto_fillable(
    *args: TagChild | TagAttrs,
    title: str | None = None,
    window_title: str | None = None,
    **kwargs: Any,
) -> Tag:
    if window_title is None and title is not None:
        window_title = title

    return page_fillable(
        None if title is None else panel_title(title),
        *args,
        title=window_title,
        **kwargs,
    )


def _page_auto_fluid(
    *args: TagChild | TagAttrs,
    title: str | None = None,
    window_title: str | None = None,
    **kwargs: str,
) -> Tag:
    if window_title is None and title is not None:
        window_title = title

    return page_fluid(
        None if title is None else panel_title(title),
        *args,
        title=window_title,
        **kwargs,
    )


def _page_auto_fixed(
    *args: TagChild | TagAttrs,
    title: str | None = None,
    window_title: str | None = None,
    **kwargs: Any,
) -> Tag:
    if window_title is None and title is not None:
        window_title = title

    return page_fixed(
        None if title is None else panel_title(title),
        *args,
        title=window_title,
        **kwargs,
    )


@no_example()
def page_output(id: str) -> Tag:
    """
    Create a page container where the entire body is a UI output.

    Parameters
    ----------
    id
        An output id.

    Returns
    -------
    :
        A UI element which is meant to be used as a page container.
    """
    return tags.html(
        tags.head(),
        tags.body(id=id, class_="shiny-page-output"),
        page_output_dependency(),
        lang="en",
    )



from __future__ import annotations

__all__ = ("Progress",)

from types import TracebackType
from typing import TYPE_CHECKING, Optional, Type
from warnings import warn

from .._docstring import add_example
from .._utils import rand_hex
from ..session import require_active_session
from ..session._session import UpdateProgressMessage

if TYPE_CHECKING:
    from ..session import Session


@add_example()
class Progress:
    """
    Initialize a progress bar.

    `Progress` creates a computation manager that can be used with `with` to
    run a block of code. Shiny will display a progress bar while the code runs, which
    you can update by calling the `set()` and `message()` methods of the computation
    manager at strategic points in the code block.

    Parameters
    ----------
    min
        The value that represents the starting point of the progress bar. Must be less
        than ``max``.
    max
        The value that represents the end of the progress bar. Must be greater than
        ``min``.
    session
        The :class:`~shiny.Session` instance that the progress bar should appear in. If not
        provided, the session is inferred via :func:`~shiny.session.get_current_session`.
    """

    _style = "notification"

    min: int
    max: int
    value: float | None

    def __init__(
        self, min: int = 0, max: int = 1, session: Optional[Session] = None
    ) -> None:
        self.min = min
        self.max = max
        self.value = None
        self._closed = False
        self._session = require_active_session(session)
        self._id = self._session.ns(rand_hex(8))

        self._session._send_progress("open", {"id": self._id, "style": self._style})

    def __enter__(self) -> "Progress":
        return self

    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> None:
        self.close()

    def set(
        self,
        value: Optional[float] = None,
        message: Optional[str] = None,
        detail: Optional[str] = None,
    ) -> None:
        """
        Opens and updates the progress panel.

        When called the first time, the progress panel is displayed.

        Parameters
        ----------
        self
            The object instance
        value
            The value at which to set the progress bar, relative to ``min`` and ``max``.
            ``None`` hides the progress bar, if it is currently visible.
        message
            The message to be displayed to the user or ``None`` to hide the current
            message (if any).
        detail
            The detail message to be displayed to the user or ``None`` to hide the
            current detail message (if any). The detail message will be shown with a
            de-emphasized appearance relative to message.
        """

        if self._closed:
            warn("Attempting to set progress, but progress already closed.")
            return None

        self.value = value
        if value:
            # Normalize value to number between 0 and 1
            value = min(1, max(0, (value - self.min) / (self.max - self.min)))

        msg: UpdateProgressMessage = {
            "id": self._id,
            "style": self._style,
        }
        if message is not None:
            msg["message"] = message
        if detail is not None:
            msg["detail"] = detail
        if value is not None:
            msg["value"] = value

        self._session._send_progress("update", msg)

    def inc(
        self,
        amount: float = 0.1,
        message: Optional[str] = None,
        detail: Optional[str] = None,
    ) -> None:
        """
        Increment the progress bar.

        Like ``set``, this updates the progress panel. The difference is that ``inc``
        increases the progress bar by amount, instead of setting it to a specific value.

        Parameters
        ----------
        self
            The object instance
        amount
            The amount to increment in progress.
        message
            The message to be displayed to the user or ``None`` to hide the current
            message (if any).
        detail
            The detail message to be displayed to the user or ``None`` to hide the current
            detail message (if any). The detail message will be shown with a
            de-emphasized appearance relative to message.
        """

        if self.value is None:
            self.value = self.min

        value = min(self.value + amount, self.max)
        self.set(value, message, detail)

    def close(self) -> None:
        """
        Close the progress bar. You can also use the Progress object as a context
        manager, which will cause the progress bar to close on exit.

        Parameters
        ----------
        self
            The object instance

        Note
        ----
        Removes the progress panel. Future calls to set and close will be ignored.
        """
        if self._closed:
            warn("Attempting to close progress, but progress already closed.")
            return None

        self._session._send_progress("close", {"id": self._id, "style": self._style})
        self._closed = True



from __future__ import annotations

import os
import pathlib
import re
import tempfile
import textwrap
from typing import TYPE_CHECKING, Any, Literal, Optional, Sequence, TypeVar

if TYPE_CHECKING:
    from brand_yml import Brand
from htmltools import HTMLDependency

from .._docstring import add_example
from .._typing_extensions import NotRequired, TypedDict
from .._versions import bootstrap
from ._theme_presets import (
    ShinyThemePreset,
    shiny_theme_presets,
    shiny_theme_presets_bundled,
)
from ._utils import path_pkg_www

T = TypeVar("T", bound="Theme")


class SassCompileArgs(TypedDict):
    output_style: NotRequired[Literal["nested", "expanded", "compact", "compressed"]]
    source_comments: NotRequired[bool]
    source_map_contents: NotRequired[bool]
    source_map_embed: NotRequired[bool]
    omit_source_map_url: NotRequired[bool]
    source_map_root: NotRequired[str | None]
    include_paths: NotRequired[Sequence[str]]
    precision: NotRequired[int]
    custom_functions: NotRequired[Any]  # not worth the effort, it's a complicated type
    indented: NotRequired[bool]
    importers: NotRequired[Any]  # not worth the effort, it's a complicated type


theme_temporary_directories: set[tempfile.TemporaryDirectory[str]] = set()


@add_example()
class Theme:
    """
    Create a custom Shiny theme.

    The `Theme` class allows you to create a custom Shiny theme by providing custom Sass
    code. The theme can be based on one of the available presets, such as `"shiny"` or
    `"bootstrap"`, or a Bootswatch theme. Use the `.add_*()` methods can be chained
    together to add custom Sass functions, defaults, mixins, and rules.

    Pass the `Theme` object directly to the `theme` argument of any Shiny page function,
    such as :func:`~shiny.ui.page_sidebar` or :func:`~shiny.ui.page_navbar`. In Shiny
    Express apps, use the `theme` argument of :func:`~shiny.express.ui.page_opts` to set
    the app theme.

    **Note: Compiling custom themes requires the
    [libsass](https://pypi.org/project/libsass/) package**, which is not installed by
    default with Shiny. Use `pip install libsass` or `pip install "shiny[theme]"` to
    install it.

    Customized themes are compiled to CSS when the theme is used. The `Theme` class
    caches the compiled CSS so that it's only compiled for the first user to load your
    app, but you can speed up app loading (and avoid the runtime `libsass` dependency)
    by pre-compiling the theme CSS and saving it to a file. To do this, use the
    `.to_css()` method to render the theme to a single minified CSS string.

    ```{.python filename="my_theme.py"}
    from pathlib import Path

    from shiny import ui

    my_theme = (
        ui.Theme("shiny")
        .add_defaults(
            my_purple="#aa00aa",
        )
        .add_mixins(
            headings_color="$my-purple",
        )
    )

    with open(Path(__file__).parent / "my_theme.css", "w") as f:
        f.write(my_theme.to_css())
    ```

    Run this script with `python my_theme.py` to generate the CSS file. Once saved to a
    file, the CSS can be used in any Shiny app by passing the file path to the `theme`
    argument instead of the `Theme` object.

    ```{.python filename="app.py"}
    from pathlib import Path

    from shiny import App, ui

    app_ui = ui.page_fluid(
        ui.h2("Hello, themed Shiny!"),
        # App content here
        title="My App",
        theme=Path(__file__).parent / "my_theme.css",
    )

    def server(input):
        pass

    app = App(app_ui, server)
    ```

    Parameters
    ----------
    preset
        The name of the preset to use as a base. `"shiny"` is the default theme for
        Shiny apps and `"bootstrap"` uses standard Bootstrap 5 styling. Bootswatch theme
        presets are also available. Use `Theme.available_presets()` to see the full
        list.
    name
        A custom name for the theme. If not provided, the preset name will be used.
    include_paths
        Additional paths to include when looking for Sass files used in `@import`
        statements in the theme. This can be a single path as a string or
        :class:`pathlib.Path`, or a list of paths. The paths should point to directories
        containing additional Sass files that the theme depends on.

    Raises
    ------
    ValueError
        If the `preset` is not a valid theme preset.
    """

    def __init__(
        self,
        preset: str | None = None,
        name: Optional[str] = None,
        include_paths: Optional[str | pathlib.Path | list[str | pathlib.Path]] = None,
    ):
        self._preset: ShinyThemePreset = check_is_valid_preset(preset or "shiny")
        self.name = name
        # 2024-06-21: `version` is not exposed because we currently support only BS 5.
        # In the future, the Bootstrap version could be chosen by the user on init.
        self._version = bootstrap
        self._include_paths: list[str] = []

        if isinstance(include_paths, (str, pathlib.Path)):
            self._include_paths.append(str(include_paths))
        elif isinstance(include_paths, Sequence):
            for path in include_paths:
                self._include_paths.append(str(path))

        # User-provided Sass code
        self._functions: list[str] = []
        self._defaults: list[str] = []
        self._mixins: list[str] = []
        self._rules: list[str] = []

        # _css is either:
        # 1. "" indicating that the CSS has not been compiled yet
        # 2. A string containing the compiled CSS for the current theme
        self._css: str = ""

        # If the theme has been customized and rendered once, we store the tempdir
        # so that we can re-use the already compiled CSS file.
        self._css_temp_srcdir: Optional[tempfile.TemporaryDirectory[str]] = None

    @staticmethod
    def available_presets() -> tuple[ShinyThemePreset, ...]:
        """
        Get a list of available theme presets.
        """
        return shiny_theme_presets

    @property
    def preset(self) -> ShinyThemePreset:
        return self._preset

    @preset.setter
    def preset(self, value: ShinyThemePreset) -> None:
        check_is_valid_preset(value)
        self._preset = value
        self._reset_css()

    def _reset_css(self) -> None:
        self._css = ""
        if self._css_temp_srcdir is not None:
            self._css_temp_srcdir.cleanup()
            theme_temporary_directories.discard(self._css_temp_srcdir)
        self._css_temp_srcdir = None

    def _get_css_tempdir(self) -> str:
        """
        Get or create a temporary directory for storing compiled CSS.

        Creates a directory via `tempfile.TemporaryDirectory` that is cleaned up
        automatically when any references to the directory are removed. When a `Theme()`
        is created and passed directly to the `theme` argument of a Shiny page function,
        the UI is rendered and the theme object, no longer needed, is garbage collected.
        To avoid cleaning up the temporary directory before its files are served, we
        keep a reference to the directory in the `theme_temporary_directories` set.
        """
        if self._css_temp_srcdir is not None:
            return self._css_temp_srcdir.name

        tempdir = tempfile.TemporaryDirectory()
        theme_temporary_directories.add(tempdir)
        self._css_temp_srcdir = tempdir
        return tempdir.name

    def _has_customizations(self) -> bool:
        return (
            len(self._functions) > 0
            or len(self._defaults) > 0
            or len(self._mixins) > 0
            or len(self._rules) > 0
        )

    def _can_use_precompiled(self) -> bool:
        return (
            self._preset in shiny_theme_presets_bundled
            and not self._has_customizations()
        )

    @staticmethod
    def _combine_args_kwargs(
        *args: str,
        allow_both: bool = True,
        kwargs: dict[str, str | float | int | bool | None],
        is_default: bool = False,
    ) -> list[str]:
        if not allow_both and len(args) > 0 and len(kwargs) > 0:
            # Python forces positional arguments to come _before_ kwargs, but default
            # argument order might matter. To be safe, we force users to pick one order.
            raise ValueError("Cannot provide both positional and keyword arguments.")
        if len(args) == 0 and len(kwargs) == 0:
            return []

        values: list[str] = list(args)
        default = " !default" if is_default else ""

        if len(kwargs) > 0:
            for key, value in kwargs.items():
                key = key.replace("_", "-")
                if isinstance(value, bool):
                    value = "true" if value else "false"
                elif value is None:
                    value = "null"
                values.append(f"${key}: {value}{default};")

        return [textwrap.dedent(x) for x in values]

    def add_functions(self: T, *args: str) -> T:
        """
        Add custom Sass functions to the theme.

        Sass code added via this method will be placed **after** the function
        declarations from the theme preset, allowing you to override or extend the
        default functions.

        Parameters
        ----------
        *args
            The Sass functions to add as a single or multiple strings.
        """
        functions = self._combine_args_kwargs(*args, kwargs={})
        self._functions.extend(functions)
        self._reset_css()
        return self

    def add_defaults(
        self: T,
        *args: str,
        **kwargs: str | float | int | bool | None,
    ) -> T:
        """
        Add custom default values to the theme.

        Sass code added via this method will be placed **before** the default values of
        the theme preset, allowing you to override or extend the default values.

        Parameters
        ----------
        *args
            Sass code, as a single or multiple strings, containing default value
            declarations to add.
        **kwargs
            Keyword arguments containing default value declarations to add. The keys
            should be Sass variable names using underscore casing that will be
            transformed automatically to kebab-case. For example,
            `.add_defaults(primary_color="#ff0000")` is equivalent to
            `.add_defaults("$primary-color: #ff0000 !default;")`.
        """
        defaults = self._combine_args_kwargs(
            *args,
            kwargs=kwargs,
            allow_both=False,
            is_default=True,
        )

        # Add args to the front of _defaults
        self._defaults[:0] = defaults
        self._reset_css()

        return self

    def add_mixins(
        self: T,
        *args: str,
        **kwargs: str | float | int | bool | None,
    ) -> T:
        """
        Add custom Sass mixins to the theme.

        Sass code added via this method will be placed **after** the mixin declarations
        from the theme preset, allowing you to override or extend the default mixins.

        Parameters
        ----------
        *args
            Sass code, as a single or multiple strings, containing mixins to add.
        **kwargs
            Keyword arguments containing Sass value declarations to add. The keys
            should be Sass variable names using underscore casing that will be
            transformed automatically to kebab-case. For example,
            `.add_mixins(primary_color="#ff0000")` is equivalent to
            `.add_mixins("$primary-color: #ff0000;")`.
        """
        mixins = self._combine_args_kwargs(*args, kwargs=kwargs)
        self._mixins.extend(mixins)
        self._reset_css()
        return self

    def add_rules(
        self: T,
        *args: str,
        **kwargs: str | float | int | bool | None,
    ) -> T:
        """
        Add custom Sass rules to the theme.

        Sass code added via this method will be placed **after** the rule declarations
        from the theme preset, allowing you to override or extend the default rules.

        Parameters
        ----------
        *args
            Sass code, as a single or multiple strings, containing rules to add.
        **kwargs
            Keyword arguments containing Sass value declarations to add. The keys
            should be Sass variable names using underscore casing that will be
            transformed automatically to kebab-case. For example,
            `.add_rules(primary_color="#ff0000")` is equivalent to
            `.add_rules("$primary-color: #ff0000;")`.
        """
        rules = self._combine_args_kwargs(*args, kwargs=kwargs)
        self._rules.extend(rules)
        self._reset_css()
        return self

    def to_sass(self) -> str:
        """
        Returns the custom theme as a single Sass string.

        Returns
        -------
        :
            The custom theme as a single Sass string.
        """
        path_functions = path_pkg_preset(self._preset, "_01_functions.scss")
        path_defaults = path_pkg_preset(self._preset, "_02_defaults.scss")
        path_mixins = path_pkg_preset(self._preset, "_03_mixins.scss")
        path_rules = path_pkg_preset(self._preset, "_04_rules.scss")

        sass_lines = [
            f'@import "{path_functions}";',
            *self._functions,
            *self._defaults,
            f'@import "{path_defaults}";',
            f'@import "{path_mixins}";',
            *self._mixins,
            f'@import "{path_rules}";',
            *self._rules,
        ]

        return "\n".join(sass_lines)

    def to_css(
        self,
        compile_args: Optional[SassCompileArgs] = None,
    ) -> str:
        """
        Compile the theme to CSS and return the result as a string.

        Parameters
        ----------
        compile_args
            A dictionary of keyword arguments to pass to
            [`sass.compile()`](https://sass.github.io/libsass-python/sass.html#sass.compile).

        Returns
        -------
        :
            The compiled CSS for the theme. The value is cached such that previously
            compiled themes are returned immediately. Adding additional custom Sass code
            or changing the preset will invalidate the cache.
        """
        if self._css:
            return self._css

        if self._can_use_precompiled():
            self._css = self._read_precompiled_css()
            return self._css

        check_theme_pkg_installed("libsass", "sass")
        import sass

        args: SassCompileArgs = {} if compile_args is None else compile_args

        if "include_paths" in args:
            raise ValueError(
                "The 'include_paths' argument is not allowed in 'compile_args'. "
                "Use the 'include_paths' argument of the Theme constructor instead.",
            )

        args: SassCompileArgs = {
            "output_style": "compressed",
            "include_paths": self._include_paths,
            **args,
        }

        self._css = sass.compile(string=self.to_sass(), **args)

        return self._css

    # Third party theme-providers, e.g. shinyswatch, can override the next three methods
    # to customize the HTML Dependency object that is returned by the theme or to
    # provide pre-compiled CSS files.
    def _dep_name(self) -> str:
        """A method returning the name of the HTML dependency."""
        return f"shiny-theme-{self.name or self._preset}"

    def _dep_css_name(self) -> str:
        """A method returning the name of the CSS file used in the HTML dependency."""
        return "bootstrap.min.css"

    def _dep_css_precompiled_path(self) -> str | pathlib.Path:
        """A method returning the path to the precompiled CSS file."""
        return path_pkg_preset(self._preset, self._dep_css_name())

    def _read_precompiled_css(self) -> str:
        path = self._dep_css_precompiled_path()
        with open(path, "r") as f:
            return f.read()

    def _dep_create(self, css_path: str | pathlib.Path) -> HTMLDependency:
        css_path = pathlib.Path(css_path)
        return HTMLDependency(
            name=make_valid_path_str(self._dep_name()),
            version=self._version,
            source={"subdir": str(css_path.parent)},
            stylesheet={
                "href": css_path.name,
                "data-shiny-theme": self.name or self._preset,  # type: ignore
            },
            # Branded themes re-use this tempdir
            all_files=False,
        )

    def _html_dependency_precompiled(self) -> HTMLDependency:
        return self._dep_create(css_path=self._dep_css_precompiled_path())

    def _html_dependencies(self) -> list[HTMLDependency]:
        """
        Create an `HTMLDependency` object from the theme.

        Returns
        -------
        :
            An list of :class:`~htmltools.HTMLDependency` objects representing the
            theme. In most cases, you should not need to call this method directly.
            Instead, pass the `Theme` object directly to the `theme` argument of any
            Shiny page function.
        """
        # Note: return a list so that this method can be overridden in subclasses of
        # Theme that want to attach additional dependencies to the theme dependency.
        if self._can_use_precompiled():
            return [self._html_dependency_precompiled()]

        css_name = self._dep_css_name()
        css_path = os.path.join(self._get_css_tempdir(), css_name)

        if not os.path.exists(css_path):
            with open(css_path, "w") as css_file:
                css_file.write(self.to_css())

        return [self._dep_create(css_path)]

    def tagify(self) -> None:
        raise SyntaxError(
            "The `Theme` class is not meant to be used as a standalone HTML tag. "
            "Instead, pass the `Theme` object directly to the `theme` argument of "
            "`shiny.express.ui.page_opts()` (Shiny Express) "
            "or any `shiny.ui.page_*()` function (Shiny Core)."
        )

    @classmethod
    def from_brand(cls, brand: "str | pathlib.Path | Brand"):
        """
        Create a custom Shiny theme from a `_brand.yml`

        Creates a custom Shiny theme for your brand using
        [brand.yml](https://posit-dev.github.io/brand-yml), a single YAML file that
        describes the brand's color and typography. Learn more about writing a
        `_brand.yml` file for your brand at the
        [brand.yml homepage](https://posit-dev.github.io/brand-yml).

        As a simple example, suppose your brand guidelines include a color palette with
        custom orange and black colors. The orange is used as the primary accent color
        and the black for all text. For typography, the brand also uses
        [Roboto](https://fonts.google.com/specimen/Roboto?query=roboto) and
        [Roboto Mono](https://fonts.google.com/specimen/Roboto+Mono?query=roboto) from
        Google Fonts for text and monospace-styled text, respectively. Here's a
        `_brand.yml` file for this brand:

        ```{.yaml filename="_brand.yml"}
        meta:
          name: brand.yml Example

        color:
          palette:
            orange: "#F96302"
            black: "#000000"
          foreground: black
          primary: orange

        typography:
          fonts:
            - family: Roboto
              source: google
            - family: Roboto Mono
              source: google
          base: Roboto
          monospace: Roboto Mono
        ```

        You can store the `_brand.yml` file next to your Shiny `app.py` or, for larger
        projects, in a parent folder. To use a theme generated from the `_brand.yml`
        file, call :meth:`~shiny.ui.Theme.from_brand` on `__file__` and pass the result
        to the `theme` argument of :func:`~shiny.express.ui.page_opts` (Shiny Express)
        or the `theme` argument of `shiny.ui.page_*` functions, like
        :func:`~shiny.ui.page_sidebar`.

        ```{.python filename="app.py"}
        from shiny.express import input, render, ui

        ui.page_opts(theme=ui.Theme.from_brand(__file__))

        ui.input_slider("n", "N", 0, 100, 20)


        @render.code
        def txt():
            return f"n*2 is {input.n() * 2}"
        ```

        Parameters
        ----------
        brand
            A :class:`brand_yml.Brand` instance, or a path to help locate `_brand.yml`.
            For a path, you can pass `__file__` or a directory containing the
            `_brand.yml` or a path directly to the `_brand.yml` file.

        Returns
        -------
        :
            A :class:`shiny.ui.Theme` instance with a custom Shiny theme created from
            the brand guidelines (see :class:`brand_yml.Brand`).
        """
        check_theme_pkg_installed("brand_yml")

        from brand_yml import Brand

        from ._theme_brand import ThemeBrand  # avoid circular import

        if not isinstance(brand, Brand):
            brand = Brand.from_yaml(brand)

        return ThemeBrand(brand)


def path_pkg_preset(preset: ShinyThemePreset, *args: str) -> str:
    """
    Returns a path relative to the packaged directory for a given preset.

    Examples
    --------

    ```python
    path_pkg_preset("shiny", "bootstrap.min.css")
    #> "{shiny}/www/shared/sass/preset/shiny/bootstrap.min.css"
    ```
    """
    path = os.path.realpath(path_pkg_www("sass", "preset", str(preset), *args))
    # these paths end up in strings passed to sass.compile(). Converting to POSIX means
    # we use forward slashes to avoid needing to double-escape the Windows backslash
    return pathlib.Path(path).as_posix()


def check_is_valid_preset(preset: str) -> ShinyThemePreset:
    if preset not in shiny_theme_presets:
        raise ValueError(
            f"Invalid preset '{preset}'.\n"
            + f"""Expected one of: "{'", "'.join(shiny_theme_presets)}".""",
        )

    return preset


def check_theme_pkg_installed(pkg: str, spec: str | None = None) -> None:
    import importlib.util

    if importlib.util.find_spec(spec or pkg) is None:
        raise ImportError(
            f"The '{pkg}' package is required to compile custom themes. "
            'Please install it with `pip install {pkg}` or `pip install "shiny[theme]"`.',
        )


def make_valid_path_str(x: str) -> str:
    return re.sub(r"[^a-zA-Z0-9_.-]+", "-", x).lower()



import seaborn as sns

# Import data from shared.py
from shared import df

from shiny.express import input, render, ui

ui.page_opts(title="Hello sidebar!")

with ui.sidebar():
    ui.input_select("var", "Select variable", choices=["bill_length_mm", "body_mass_g"])
    ui.input_switch("species", "Group by species", value=True)
    ui.input_switch("show_rug", "Show Rug", value=True)


@render.plot
def hist():
    hue = "species" if input.species() else None
    sns.kdeplot(df, x=input.var(), hue=hue)
    if input.show_rug():
        sns.rugplot(df, x=input.var(), hue=hue, color="black", alpha=0.25)



from shiny.express import input, render, ui

ui.input_slider("n", "N", 1, 100, 50)


# @render.express is like @render.ui, but with Express syntax
@render.express
def render_express1():
    "Slider value:"
    input.n()


# @render.express() also works with parens
@render.express()
def render_express2():
    "Slider value:"
    input.n()



import datetime
import re
import time

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc
from shiny.types import MISSING


def convert_to_utc_date(date_str: str) -> str:
    date_obj = datetime.datetime.strptime(date_str, "%m/%d/%y")
    epoch_time_milliseconds = date_obj.timestamp() * 1000
    return str(epoch_time_milliseconds)


def convert_to_utc_date_time(date_time_str: str) -> str:
    date_time_obj = datetime.datetime.strptime(date_time_str, "%Y-%m-%d %H:%M:%S")
    utc_timestamp = (
        date_time_obj.replace(tzinfo=datetime.timezone.utc).timestamp() * 1000
    )
    return str(utc_timestamp)


def test_slider_regular(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s0 = controller.InputSlider(page, "s0")
    s0.expect_label("Integer")
    s0.expect_value("500")
    s0.expect_min("0")
    s0.expect_max("1000")
    s0.expect_step("1")
    s0.expect_ticks("false")
    s0.expect_sep(",")
    s0.expect_pre(None)
    s0.expect_post(None)
    s0.expect_time_format(None)
    s0.expect_timezone(None)
    s0.expect_drag_range(None)
    s0.expect_animate(exists=False)
    controller.OutputTextVerbatim(page, "txt0").expect_value("500")

    new_val = "20"
    s0.set(new_val)
    s0.expect_value(new_val)
    controller.OutputTextVerbatim(page, "txt0").expect_value(new_val)


def test_slider_range(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s1 = controller.InputSliderRange(page, "s1")
    s1.expect_label("Range")
    s1.expect_value(("200", "500"))
    s1.expect_min("1")
    s1.expect_max("1000")
    s1.expect_step("1")
    s1.expect_ticks("false")
    s1.expect_sep(",")
    s1.expect_pre(None)
    s1.expect_post(None)
    s1.expect_time_format(None)
    s1.expect_timezone(None)
    s1.expect_drag_range("true")
    s1.expect_animate(exists=False)
    controller.OutputTextVerbatim(page, "txt1").expect_value("(200, 500)")

    new_val = ("605", "840")
    s1.set(new_val, max_err_values=1000)
    try:
        s1.expect_value((MISSING, MISSING))  # type: ignore
    except ValueError as e:
        assert re.search("tuple entries cannot", str(e))
    s1.expect_value((new_val[0], MISSING))
    s1.expect_value((MISSING, new_val[1]))
    s1.expect_value(new_val)
    controller.OutputTextVerbatim(page, "txt1").expect_value(
        f"({new_val[0]}, {new_val[1]})"
    )


def test_slider_custom_format(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s2 = controller.InputSlider(page, "s2")
    s2.expect_label("Custom Format")
    s2.expect_value("$0.00")
    s2.expect_min("0")
    s2.expect_max("10000")
    s2.expect_step("2500")
    s2.expect_ticks("true")
    s2.expect_sep(",")
    s2.expect_pre("$")
    s2.expect_post(".00")
    s2.expect_time_format(None)
    s2.expect_timezone(None)
    s2.expect_drag_range(None)
    s2.expect_animate_options(loop=True, interval=500)
    controller.OutputTextVerbatim(page, "txt2").expect_value("0")

    s2.set("$7,500.00")
    s2.expect_value("$7,500.00")
    controller.OutputTextVerbatim(page, "txt2").expect_value("7500")


def test_slider_loop(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s3 = controller.InputSlider(page, "s3")
    s3.expect_label("Looping Animation")
    s3.expect_value("1,000")
    s3.expect_min("1")
    s3.expect_max("2000")
    s3.expect_step("10")
    s3.expect_ticks("false")
    s3.expect_sep(",")
    s3.expect_pre(None)
    s3.expect_post(None)
    s3.expect_time_format(None)
    s3.expect_timezone(None)
    s3.expect_drag_range(None)
    s3.expect_animate_options(loop=True, interval=300)
    controller.OutputTextVerbatim(page, "txt3").expect_value("1000")

    s3.set("1,441")
    s3.expect_value("1,441")
    controller.OutputTextVerbatim(page, "txt3").expect_value("1441")

    # Play for a little bit
    s3.click_play()
    time.sleep(400 / 1000)
    s3.click_pause()
    # Make sure the value changed
    try:
        s3.expect_value("1,441", timeout=100)
    except AssertionError:
        pass


def test_slider_play(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s4 = controller.InputSlider(page, "s4")
    s4.expect_label("Single Animation")
    s4.expect_value("1")
    s4.expect_min("0")
    s4.expect_max("5")
    s4.expect_step("1")
    s4.expect_ticks("false")
    s4.expect_sep(",")
    s4.expect_pre(None)
    s4.expect_post(None)
    s4.expect_time_format(None)
    s4.expect_timezone(None)
    s4.expect_drag_range(None)
    s4.expect_animate_options(loop=False, interval=100)
    controller.OutputTextVerbatim(page, "txt4").expect_value("1")

    s4.click_play()
    s4.expect_value("5")
    s4.click_play()  # can click again!
    s4.expect_value("5")


def test_slider_date_format(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s5 = controller.InputSlider(page, "s5")
    s5.expect_label("Date format")
    s5.expect_value("01/05/24")
    s5.expect_min(convert_to_utc_date("01/01/24"))
    s5.expect_max(convert_to_utc_date("01/10/24"))
    s5.expect_time_format("%m/%d/%y")
    s5.expect_timezone("0000")
    s5.expect_drag_range(None)
    controller.OutputTextVerbatim(page, "txt5").expect_value("2024-01-05")

    new_val = "01/08/24"
    s5.set(new_val)
    s5.expect_value(new_val)
    controller.OutputTextVerbatim(page, "txt5").expect_value("2024-01-08")


def test_slider_time_format(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s6 = controller.InputSlider(page, "s6")
    s6.expect_label("Time format")
    s6.expect_timezone("0000")
    s6.expect_value("2024-01-05 12:00:00")
    s6.expect_min(convert_to_utc_date_time("2024-01-01 00:00:00"))
    s6.expect_max(convert_to_utc_date_time("2024-01-10 23:59:00"))
    s6.expect_time_format("%F %T")
    s6.expect_width("600px")
    s6.expect_drag_range(None)
    controller.OutputTextVerbatim(page, "txt6").expect_value("2024-01-05 12:00:00")

    new_val = "2024-01-01 00:00:00"
    s6.set(new_val)
    s6.expect_value(new_val)
    controller.OutputTextVerbatim(page, "txt6").expect_value("2024-01-01 00:00:00")


def test_slider_drag_range_disabled(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    s7 = controller.InputSliderRange(page, "s7")
    s7.expect_label("Drag Range (Disabled)")
    s7.expect_value(("200", "500"))
    s7.expect_min("0")
    s7.expect_max("1000")
    s7.expect_drag_range("false")
    new_val = ("25", "502")
    s7.set(new_val, max_err_values=1000)
    controller.OutputTextVerbatim(page, "txt7").expect_value(
        f"({new_val[0]}, {new_val[1]})"
    )



from shiny import App, ui

app_ui = ui.page_fluid(
    ui.card(
        ui.card_header("Check for header"),
        "This is the body of a card with default height w/ fullscreen",
        ui.card_footer("Check for footer"),
        full_screen=True,
        id="card1",
    ),
    ui.card(
        ui.p("This is the body without a header of a footer - No Fullscreen"),
        full_screen=False,
        id="card2",
    ),
    ui.card(
        ui.card_header("Fill is False. Fullscreen is False"),
        ui.h3("Max height and min height are set."),
        fill=False,
        max_height="500px",
        min_height="200px",
        id="card3",
    ),
)


app = App(app_ui, server=None)



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_popover(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    popover = controller.Popover(page, "popover_id")
    popover.expect_active(False)
    popover.set(True)
    popover.expect_active(True)
    popover.expect_body("A message")
    popover.expect_placement("bottom")
    popover.set(False)
    popover.expect_active(False)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, cast, overload

from htmltools import TagNode

from ..._typing_extensions import TypeIs
from ...types import Jsonifiable
from ._types import CellHtml, Series

if TYPE_CHECKING:
    from ...session import Session
    from ...session._utils import RenderedDeps


def as_cell_html(processed_ui: RenderedDeps) -> CellHtml:
    return {"isShinyHtml": True, "obj": processed_ui}


@overload
def maybe_as_cell_html(  # pyright: ignore[reportOverlappingOverload]
    x: str, *, session: Session
) -> Jsonifiable: ...
@overload
def maybe_as_cell_html(  # pyright: ignore[reportOverlappingOverload]
    x: TagNode, *, session: Session
) -> CellHtml: ...
@overload
def maybe_as_cell_html(x: Jsonifiable, *, session: Session) -> Jsonifiable: ...
def maybe_as_cell_html(
    x: Jsonifiable | TagNode, *, session: Session
) -> Jsonifiable | CellHtml:
    if ui_must_be_processed(x):
        return as_cell_html(session._process_ui(x))
    return cast(Jsonifiable, x)


def series_contains_htmltoolslike(ser: Series) -> bool:

    for idx, val in enumerate(ser):
        # Only check the first 1k elements in the series
        # This could lead to false negatives, but it's a reasonable tradeoff of speed.
        # If the user has _that_ much missing data and wants HTML support,
        #   they can preprocess their data to turn `None` values into `htmltools.HTML("")`
        if idx > 1000:
            return False

        if val is None:
            continue

        # Q: Can we short circuit this any quicker?
        # A: Not really. We need to check every element in the series
        #    as the typical dtype is greedy does not enforce a single type.
        # Reprex:
        # pl.Series([{"y": 2}, {"x": 1}, None, HTML("<p>Hello</p>")]).dtype
        # #> Struct({'y': Int64})
        if ui_must_be_processed(val):
            return True
    return False


@overload
def ui_must_be_processed(  # pyright: ignore[reportOverlappingOverload]
    val: str,
) -> Literal[False]: ...
@overload
def ui_must_be_processed(
    val: TagNode | object,
) -> TypeIs[TagNode]: ...
def ui_must_be_processed(  # pyright: ignore[reportInconsistentOverload]
    val: object,
):
    from htmltools import is_tag_node

    if isinstance(val, str):
        return False

    return is_tag_node(val)



from __future__ import annotations

from shiny import App, ui
from shiny.types import NavSetArg

my_sidebar = ui.sidebar("Sidebar content", open="open", title="Sidebar title")


def nav_with_content(letter: str, prefix: str) -> ui._navs.NavPanel:
    return ui.nav_panel(letter, ui.markdown(f"`{prefix}`: tab {letter} content"))


def nav_items(prefix: str) -> list[NavSetArg]:
    a = nav_with_content("a", prefix)
    b = nav_with_content("b", prefix)
    c = nav_with_content("c", prefix)
    space = ui.nav_spacer()
    links = ui.nav_menu(
        "Links",
        ui.nav_control(
            ui.tags.a(
                # ui.icon("github"),
                "Shiny",
                href="https://github.com/posit-dev/py-shiny",
                target="_blank",
            ),
        ),
        "---",
        "Plain text",
        "---",
        ui.nav_control(
            ui.a(
                "Posit",
                href="https://posit.co",
                target="_blank",
            )
        ),
        align="right",
    )
    return [a, b, c, space, links]


app = App(
    ui=ui.page_navbar(
        # theme = bs_theme(),
        *nav_items("page_navbar()"),
        sidebar=my_sidebar,
        title="page_navbar()",
        bg="#0062cc",
        header=ui.markdown(
            "Testing app for `bslib::nav_spacer()` and `bslib::nav_item()` [#319](https://github.com/rstudio/bslib/pull/319)."
        ),
        footer=ui.div(
            {"style": "width:80%; margin: 0 auto"},
            ui.h4("navset_card_tab()"),
            ui.navset_card_tab(
                *nav_items("navset_card_tab()"),
                sidebar=my_sidebar,
            ),
            ui.h4("navset_card_pill()"),
            ui.navset_card_pill(
                *nav_items("navset_card_pill()"),
                sidebar=my_sidebar,
            ),
            # Do not include `navset_bar()` in example. Ok for testing only
            ui.h4("navset_bar()"),
            ui.navset_bar(
                *nav_items("navset_bar()"),
                title="Test!",
                sidebar=my_sidebar,
            ),
        ),
    ),
    server=None,
)



from shiny import App, Inputs, Outputs, Session, reactive, ui


def make_panel(letter: str) -> ui.AccordionPanel:
    return ui.accordion_panel(
        f"Section {letter}",
        f"Some narrative for section {letter}",
        value=f"sec_{letter}",
    )


items = [make_panel(letter) for letter in "ABCDE"]

app_ui = ui.page_fluid(
    ui.input_switch("update_panel", "Update (and open) Sections"),
    ui.accordion(*items, id="acc", multiple=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.update_panel)
    def _():
        txt = " (updated)" if input.update_panel() else ""
        show = bool(input.update_panel() % 2 == 1)
        for letter in "ABCDE":
            ui.update_accordion_panel(
                "acc",
                f"sec_{letter}",
                f"Some{txt} narrative for section {letter}",
                title=f"Section {letter}{txt}",
                # Open Accordion Panel to see updated contents
                show=show,
            )
        next_show_txt = "close" if show else "open"

        ui.update_switch("update_panel", label=f"Update (and {next_show_txt}) Sections")


app = App(app_ui, server)



from typing import Any, Dict

from shiny import ui as core_ui
from shiny.express import expressify, ui

ui.page_opts(title="Navsets kitchensink App", id="navsets_collection")


def navset_sidebar():
    return core_ui.sidebar(core_ui.markdown("Sidebar content"))


navset_configs: Dict[str, Dict[str, Dict[str, Any]]] = {
    "navset_pill": {
        "default": {},
        "with_header_footer": {
            "header": {
                "content": "navset_pill_with_header_footer header",
                "id": "navset_pill_header",
            },
            "footer": {
                "content": "navset_pill_with_header_footer footer",
                "id": "navset_pill_footer",
            },
        },
        "selected": {"selected": "navset_pill_b"},
    },
    "navset_underline": {
        "default": {},
        "with_header_footer": {
            "header": {
                "content": "navset_underline_with_header_footer header",
                "id": "navset_underline_header",
            },
            "footer": {
                "content": "navset_underline_with_header_footer footer",
                "id": "navset_underline_footer",
            },
        },
        "selected": {"selected": "navset_underline_b"},
    },
    "navset_tab": {
        "default": {},
        "with_header_footer": {
            "header": {
                "content": "navset_tab_with_header_footer header",
                "id": "navset_tab_header",
            },
            "footer": {
                "content": "navset_tab_with_header_footer footer",
                "id": "navset_tab_footer",
            },
        },
        "selected": {"selected": "navset_tab_b"},
    },
    "navset_pill_list": {
        "default": {},
        "with_header_footer": {
            "header": {
                "content": "navset_pill_list_with_header_footer header",
                "id": "navset_pill_list_header",
            },
            "footer": {
                "content": "navset_pill_list_with_header_footer footer",
                "id": "navset_pill_list_footer",
            },
        },
        "selected": {"selected": "navset_pill_list_b"},
        "widths_no_well": {"widths": (10, 2), "well": False},
    },
    "navset_card_pill": {
        "with_header_footer": {
            "title": "navset_card_pill_with_header_footer",
            "header": {
                "content": "navset_card_pill_with_header_footer header",
                "id": "navset_card_pill_header",
            },
            "footer": {
                "content": "navset_card_pill_with_header_footer footer",
                "id": "navset_card_pill_footer",
            },
        },
        "default": {},
        "placement": {"placement": "below"},
        "selected": {"selected": "navset_card_pill_b"},
        "with_sidebar": {"sidebar": navset_sidebar()},
    },
    "navset_card_tab": {
        "with_header_footer": {
            "title": "navset_card_tab_with_header_footer",
            "header": {
                "content": "navset_card_tab_with_header_footer header",
                "id": "navset_card_tab_header",
            },
            "footer": {
                "content": "navset_card_tab_with_header_footer footer",
                "id": "navset_card_tab_footer",
            },
        },
        "default": {},
        "selected": {"selected": "navset_card_tab_b"},
        "with_sidebar": {"sidebar": navset_sidebar()},
    },
    "navset_card_underline": {
        "with_header_footer": {
            "title": "navset_card_underline_with_header_footer",
            "header": {
                "content": "navset_card_underline_with_header_footer header",
                "id": "navset_card_underline_header",
            },
            "footer": {
                "content": "navset_card_underline_with_header_footer footer",
                "id": "navset_card_underline_footer",
            },
        },
        "default": {},
        "selected": {"selected": "navset_card_underline_b"},
        "placement": {"placement": "below"},
        "with_sidebar": {"sidebar": navset_sidebar()},
    },
}


@expressify
def create_navset(navset_type: str) -> None:
    navset_function = getattr(ui, navset_type)

    for navset_id, params in navset_configs[navset_type].items():
        navset_kwargs = params.copy()

        if "header" in navset_kwargs:
            header_content = navset_kwargs["header"]["content"]
            header_id = navset_kwargs["header"]["id"]
            navset_kwargs["header"] = ui.tags.div(header_content, id=f"{header_id}")

        if "footer" in navset_kwargs:
            footer_content = navset_kwargs["footer"]["content"]
            footer_id = navset_kwargs["footer"]["id"]
            navset_kwargs["footer"] = ui.tags.div(footer_content, id=f"{footer_id}")

        with navset_function(id=f"{navset_type}_{navset_id}", **navset_kwargs):
            for suffix in ["a", "b", "c"]:
                with ui.nav_panel(f"{navset_type}_{suffix}"):
                    ui.markdown(f"{navset_type}_{suffix} content")


with ui.navset_tab(id="navsets_collection"):
    for navset_type in navset_configs.keys():
        with ui.nav_panel(navset_type):
            create_navset(navset_type)



from __future__ import annotations

# pyright: reportUnknownMemberType=false, reportUnknownVariableType=false
from pathlib import Path
from typing import Callable

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from PIL import Image
from plotnine import aes, element_rect, geom_point, theme, theme_minimal
from plotnine.data import mtcars
from plotnine.ggplot import ggplot

from shiny import App, Inputs, Outputs, Session, module, render, req, ui

tips = sns.load_dataset("tips")
dpi = 150


@module.ui
def plot_ui():
    plot_outputs = [
        ui.div(
            ui.output_plot("plot_default"),
            style="width: 300px; height: 200px;",
            class_="html-fill-container",
        ),
        ui.output_plot("plot_dom_size", width="300px", height="200px"),
        ui.output_plot("plot_decorator_size", width="auto", height="auto"),
        ui.output_plot("plot_native_size", width="auto", height="auto"),
    ]

    return ui.TagList(
        [ui.div(x, class_="border mb-3") for x in plot_outputs],
    )


@module.server
def plot_server(
    input: Inputs,
    output: Outputs,
    session: Session,
    plot_fn: Callable[[tuple[float, float] | None], object],
):
    @render.plot
    def plot_default():
        return plot_fn(None)

    @render.plot
    def plot_dom_size():
        return plot_fn(None)

    @render.plot(width=300, height=200)
    def plot_decorator_size():
        return plot_fn(None)

    @render.plot(width=0, height=0)
    def plot_native_size():
        return plot_fn((300, 200))


app_ui = ui.page_navbar(
    ui.nav_panel(
        "matplotlib",
        ui.p(
            "The following four plots should all be the same size. The last one should have larger text."
        ),
        plot_ui("mpl"),
        value="mpl",
    ),
    ui.nav_panel(
        "seaborn",
        ui.p(
            "The following four plots should all be the same size. The last one should have larger text."
        ),
        plot_ui("sns"),
        value="sns",
    ),
    ui.nav_panel(
        "plotnine",
        ui.p(
            "The following four plots should all be the same size. The last one should have larger text."
        ),
        plot_ui("plotnine"),
    ),
    ui.nav_panel(
        "pil",
        ui.p("The following three images should all be the same size."),
        plot_ui("pil"),
    ),
    id="tabset",
)


def server(input: Inputs, output: Outputs, session: Session):
    def plot_with_mpl(fig_size: tuple[float, float] | None) -> object:
        fig, ax = plt.subplots(facecolor="lavender")
        X, Y = np.mgrid[-4:4, -4:4]
        ax.quiver(X, Y)

        # Background color
        ax.set_facecolor("lavender")

        if fig_size is not None:
            fig.set_dpi(150)
            fig.set_size_inches(
                fig_size[0] / fig.get_dpi(), fig_size[1] / fig.get_dpi()
            )

        return fig

    def plot_with_sns(fig_size: tuple[float, float] | None) -> object:
        kwargs = dict()
        if fig_size:
            kwargs["height"] = fig_size[1] / dpi
            kwargs["aspect"] = fig_size[0] / fig_size[1]

        # FacetGrid has an opinion about its figure size
        g = sns.FacetGrid(tips, **kwargs)  # pyright: ignore[reportUnknownArgumentType]
        g.figure.set_facecolor("lavender")
        g.map(
            sns.scatterplot,  # pyright: ignore[reportUnknownArgumentType]
            "total_bill",
            "tip",
        )
        plt.gca().set_facecolor("lavender")
        if fig_size:
            plt.gcf().set_dpi(dpi)

    def plot_with_plotnine(fig_size: tuple[float, float] | None) -> object:
        p = (
            ggplot(mtcars, aes("wt", "mpg"))
            + geom_point()
            + theme_minimal()
            + theme(
                plot_background=element_rect(fill="lavender"),
                legend_background=element_rect(fill="lavender"),
            )
        )
        if fig_size is not None:
            p = p + theme(
                figure_size=(fig_size[0] / dpi, fig_size[1] / dpi),
                plot_background=element_rect(fill="lavender"),
                dpi=dpi,
            )
        return p

    def plot_with_pil(fig_size: tuple[float, float] | None) -> object:
        req(fig_size is None)
        return Image.open(Path(__file__).parent / "bike.jpg")

    plot_server("mpl", plot_with_mpl)
    plot_server("sns", plot_with_sns)
    plot_server("plotnine", plot_with_plotnine)
    plot_server("pil", plot_with_pil)


app = App(app_ui, server)



from __future__ import annotations

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_fluid(
    ui.h2(ui.code("@reactive.event")),
    ui.input_action_button("btn_count", "Immediate Count"),
    ui.tags.br(),
    ui.tags.label("Rendered on click:"),
    ui.output_text_verbatim("txt_immediate", placeholder=True),
    ui.input_action_button("btn_trigger", "Update Count"),
    ui.tags.br(),
    ui.tags.label("Reactive event on renderer:"),
    ui.output_text_verbatim("txt_render_delayed", placeholder=True),
    ui.tags.label("Reactive event on reactive calc:"),
    ui.output_text_verbatim("txt_reactive_delayed", placeholder=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def txt_immediate():
        return input.btn_count()

    @render.text
    @reactive.event(input.btn_trigger)
    def txt_render_delayed():
        return input.btn_count()

    @reactive.calc()
    @reactive.event(input.btn_trigger)
    def delayed_btn_count() -> int:
        return input.btn_count()

    @render.text
    def txt_reactive_delayed():
        return str(delayed_btn_count())


app = App(app_ui, server)



"""Facade classes for working with Shiny inputs/outputs in Playwright"""

from __future__ import annotations

import typing
from typing import Optional

OptionalStr = Optional[str]
OptionalInt = Optional[int]
OptionalFloat = Optional[float]
OptionalBool = Optional[bool]

PatternStr = typing.Pattern[str]
PatternOrStr = typing.Union[str, PatternStr]
ListPatternOrStr = typing.Union[
    typing.List[PatternOrStr], typing.List[str], typing.List[PatternStr]
]
AttrValue = typing.Union[PatternOrStr, None]
StyleValue = typing.Union[PatternOrStr, None]

Timeout = typing.Union[float, None]



from shiny.express import input, render, ui

ui.page_opts(title="Kitchen Sink: ui.input_radio_buttons()", fillable=True)

options = ["Option A", "Option B", "Option C"]
options_dict = {"a": "Option A", "b": "Option B", "c": "Option C"}

with ui.layout_columns():
    with ui.card():
        ui.h3("Default Radio Buttons")
        ui.input_radio_buttons("default", "Default radio buttons", options)

        @render.code
        def default_txt():
            return str(input.default())

    with ui.card():
        ui.h3("With Selected Value")
        ui.input_radio_buttons("selected", "Preset value", options, selected="Option B")

        @render.code
        def selected_txt():
            return str(input.selected())

    with ui.card():
        ui.h3("Inline Layout")
        ui.input_radio_buttons("inline", "Inline layout", options, inline=True)

        @render.code
        def inline_txt():
            return str(input.inline())


with ui.layout_columns():
    with ui.card():
        ui.h3("Custom Width")
        ui.input_radio_buttons("width", "Custom width", options, width="30px")

        @render.code
        def width_txt():
            return str(input.width())

    with ui.card():
        ui.h3("With Named List")
        ui.input_radio_buttons("choices_dict", "Named list options", options_dict)

        @render.code
        def choices_dict_txt():
            return str(input.choices_dict())



from __future__ import annotations

import base64
import io
import mimetypes
from pathlib import Path, PurePath
from typing import Literal, Optional, Protocol

from htmltools import Tag, TagAttrs, TagAttrValue, Tagifiable, css, tags

from ..._docstring import add_example
from ...ui._card import CardItem, card_body
from ...ui.css import CssUnit, as_css_unit
from ...ui.fill import as_fill_item, as_fillable_container

__all__ = ("card_image",)

############################################################################


class ImgContainer(Protocol):
    """
    A callable that wraps the return value of `card_image()`. To isolate your object in a card, return a :class:`~shiny.ui.CardItem`.
    """

    def __call__(self, *args: Tag) -> Tagifiable:
        """
        Wraps the return value of `card_image()`.

        Parameters
        ----------
        *args
            The return value of `card_image()`.

        Returns
        -------
        :
            A tagifiable object, such as a :class:`~htmltools.Tag` or
            :class:`~shiny.ui.CardItem` object.
        """
        ...


@add_example()
def card_image(
    file: str | Path | PurePath | io.BytesIO | None,
    *args: TagAttrs,
    href: Optional[str] = None,
    border_radius: Literal["top", "bottom", "all", "none"] = "top",
    mime_type: Optional[str] = None,
    class_: Optional[str] = None,
    height: Optional[CssUnit] = None,
    fill: bool = True,
    width: Optional[CssUnit] = None,
    # Required so that multiple `card_images()` are not put in the same `card()`
    container: ImgContainer = card_body,
    **kwargs: TagAttrValue,
) -> Tagifiable:
    # TODO-future: Must be updated to match rstudio/bslib#1076 before moving from exp.
    """
    A card image container

    :func:`~shiny.experimental.ui.card_image` creates a general container for an image within a
    :func:`~shiny.ui.card`. This component is designed to be
    provided as a direct child to :func:`~shiny.ui.card`.

    Parameters
    ----------
    file
        A file path pointing to an image. The image will be base64 encoded and provided to
        the `src` attribute of the `<img>` tag. Alternatively, you may set this value to
        `None` and provide the `src` yourself via `*args:TagAttrs` or
        `**kwargs:TagAttrValue` (e.g., `{"src": "HOSTED_PATH_TO_IMAGE"}` or
        `src="HOSTED_PATH_TO_IMAGE"`).
    *args
        A dictionary of tag attributes that are supplied to the resolved
        :class:`~htmltools.Tag` object.
    href
        An optional URL to link to.
    border_radius
        Where to apply `border-radius` on the image.
    mime_type
        The mime type of the `file`.
    class_
        Additional CSS classes for the resolved :class:`~htmltools.Tag` object.
    height
        Any valid CSS unit (e.g., `height="200px"`). `height` will not apply when a card is made
        `full_screen`. In this case, consider setting a `height` in
        :func:`~shiny.experimental.ui.card_body`.
    fill
        Whether to allow this element to grow/shrink to fit its `card` container.
    width
        Any valid CSS unit (e.g., `width="100%"`).
    container
        Method to wrap the returned :class:`~htmltools.Tag` object. Defaults to
        :func:`~shiny.experimental.ui.card_body`.
        If :func:`~shiny.experimental.ui.card_body` is used, each image will be in separate cards. If
        the `container` method does not return a :class:`~shiny.ui.CardItem`, it
        allows for consecutive non-`CardItem` objects to be bundled into a single
        :func:`~shiny.experimental.ui.card_body` within :func:`~shiny.ui.card`.
    **kwargs
        Additional HTML attributes for the resolved :class:`~htmltools.Tag`.
    """
    src = None
    if file is not None:
        if isinstance(file, io.BytesIO):
            b64_str = base64.b64encode(file.read()).decode("utf-8")
            if mime_type is None:
                raise ValueError(
                    "`mime_type` must be provided when passing an in-memory buffer"
                )
            src = f"data:{mime_type};base64,{b64_str}"

        elif isinstance(file, (str, Path, PurePath)):
            with open(file, "rb") as img_file:
                b64_str = base64.b64encode(img_file.read()).decode("utf-8")
                if mime_type is None:
                    mime_type = mimetypes.guess_type(file)[0]
                src = f"data:{mime_type};base64,{b64_str}"

    card_class_map = {
        "all": "card-img",
        "top": "card-img-top",
        "bottom": "card-img-bottom",
    }

    image = tags.img(
        {
            "src": src,
            "class": "img-fluid",
            "style": css(
                height=as_css_unit(height),
                width=as_css_unit(width),
            ),
        },
        {"class": card_class_map.get(border_radius, None)},
        *args,
        class_=class_,
        **kwargs,
    )

    if fill:
        image = as_fill_item(image)

    if href is not None:
        image = as_fillable_container(
            as_fill_item(
                tags.a(
                    image,
                    href=href,
                )
            )
        )

    if container:
        return container(image)
    else:
        return CardItem(image)



from faicons import icon_svg

from shiny.express import input, render, ui

ui.page_opts(title="Kitchen Sink: ui.input_action_button()", fillable=True)

with ui.layout_columns():
    with ui.card():
        ui.h3("Default Action Button")
        ui.input_action_button("default", label="Default button")

        @render.code
        def default_txt():
            return f"Button clicked {input.default()} times"

    with ui.card():
        ui.h3("With Custom Width")
        ui.input_action_button("width", "Wide button", width="200px")

        @render.code
        def width_txt():
            return f"Button clicked {input.width()} times"

    with ui.card():
        ui.h3("With Icon")
        ui.input_action_button(
            "icon", "Button with icon", icon=icon_svg("trash-arrow-up")
        )

        @render.code
        def icon_txt():
            return f"Button clicked {input.icon()} times"

    with ui.card():
        ui.h3("Disabled Button")
        ui.input_action_button("disabled", "Disabled button", disabled=True)

        @render.code
        def disabled_txt():
            return f"Button clicked {input.disabled()} times"



"""
Tools for creating user interfaces including: custom components, HTML components,
layout helpers, page-level containers, and more.
"""

from htmltools import (
    HTML,
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    TagList,
    a,
    br,
    code,
    div,
    em,
    h1,
    h2,
    h3,
    h4,
    h5,
    h6,
    head_content,
    hr,
    img,
    p,
    pre,
    span,
    strong,
    tags,
)

# The css module is for internal use, so we won't re-export it.
from . import css  # noqa: F401  # pyright: ignore[reportUnusedImport]

# Expose the fill module for extended usage: ex: ui.fill.as_fill_item(x).
from . import fill

# Export busy_indicators module
from . import busy_indicators

from ._accordion import (
    AccordionPanel,
    accordion,
    accordion_panel,
    insert_accordion_panel,
    remove_accordion_panel,
    update_accordion,
    update_accordion_panel,
)
from ._bootstrap import (
    column,
    help_text,
    panel_absolute,
    panel_conditional,
    panel_fixed,
    panel_title,
    panel_well,
    row,
)
from ._card import (
    CardItem,
    card,
    card_body,
    card_footer,
    card_header,
)
from ._chat import Chat, chat_ui
from ._download_button import download_button, download_link
from ._include_helpers import include_css, include_js
from ._input_action_button import input_action_button, input_action_link
from ._input_check_radio import (
    input_checkbox,
    input_checkbox_group,
    input_radio_buttons,
    input_switch,
)
from ._input_dark_mode import input_dark_mode, update_dark_mode
from ._input_date import input_date, input_date_range
from ._input_file import input_file
from ._input_numeric import input_numeric
from ._input_password import input_password
from ._input_select import input_select, input_selectize
from ._input_slider import AnimationOptions, SliderStepArg, SliderValueArg, input_slider
from ._input_task_button import bind_task_button, input_task_button
from ._input_text import input_text, input_text_area
from ._input_update import (
    update_action_button,
    update_action_link,
    update_checkbox,
    update_checkbox_group,
    update_date,
    update_date_range,
    update_navs,
    update_numeric,
    update_popover,
    update_radio_buttons,
    update_select,
    update_selectize,
    update_slider,
    update_switch,
    update_task_button,
    update_text,
    update_text_area,
    update_tooltip,
)
from ._insert import insert_ui, remove_ui
from ._layout import layout_column_wrap
from ._layout_columns import layout_columns
from ._markdown import markdown
from ._modal import modal, modal_button, modal_remove, modal_show
from ._navs import (
    nav_control,
    nav_menu,
    nav_panel,
    nav_spacer,
    navset_bar,
    navset_card_pill,
    navset_card_tab,
    navset_card_underline,
    navset_hidden,
    navset_pill,
    navset_pill_list,
    navset_tab,
    navset_underline,
)
from ._notification import notification_remove, notification_show
from ._output import (
    output_code,
    output_image,
    output_plot,
    output_table,
    output_text,
    output_text_verbatim,
    output_ui,
)
from ._page import (
    page_auto,
    page_bootstrap,
    page_fillable,
    page_fixed,
    page_fluid,
    page_navbar,
    page_output,
    page_sidebar,
)
from ._plot_output_opts import brush_opts, click_opts, dblclick_opts, hover_opts
from ._popover import popover
from ._progress import Progress
from ._sidebar import (
    Sidebar,
    layout_sidebar,
    sidebar,
    update_sidebar,
)
from ._theme import Theme
from ._tooltip import tooltip
from ._utils import js_eval
from ._valuebox import (
    ShowcaseLayout,
    ValueBoxTheme,
    showcase_bottom,
    showcase_left_center,
    showcase_top_right,
    value_box,
    value_box_theme,
)
from .dataframe import output_data_frame

__all__ = (
    # _bootstrap
    "row",
    "column",
    "panel_well",
    "panel_conditional",
    "panel_title",
    "panel_fixed",
    "panel_absolute",
    "help_text",
    # _sidebar
    "Sidebar",
    "sidebar",
    "layout_sidebar",
    "update_sidebar",
    # _layout
    "layout_columns",
    "layout_column_wrap",
    # _card
    "CardItem",
    "card",
    "card_body",
    "card_header",
    "card_footer",
    # _chat
    "Chat",
    "chat_ui",
    # _accordion
    "AccordionPanel",
    "accordion",
    "accordion_panel",
    "insert_accordion_panel",
    "remove_accordion_panel",
    "update_accordion",
    "update_accordion_panel",
    # _download_button
    "download_button",
    "download_link",
    # _plot_output_opts
    "brush_opts",
    "click_opts",
    "dblclick_opts",
    "hover_opts",
    # _include_helpers
    "include_css",
    "include_js",
    # _input_action_button
    "input_action_button",
    "input_action_link",
    # _input_check_radio
    "input_checkbox",
    "input_checkbox_group",
    "input_switch",
    "input_radio_buttons",
    # _input_dark_mode
    "input_dark_mode",
    "update_dark_mode",
    # _input_date
    "input_date",
    "input_date_range",
    # _input_file
    "input_file",
    # _input_numeric
    "input_numeric",
    # _input_password
    "input_password",
    # _input_select
    "input_select",
    "input_selectize",
    # _input_slider
    "input_slider",
    "SliderValueArg",
    "SliderStepArg",
    "AnimationOptions",
    # _input_task_button
    "bind_task_button",
    "input_task_button",
    # _input_text
    "input_text",
    "input_text_area",
    # _input_update
    "update_action_button",
    "update_action_link",
    "update_checkbox",
    "update_switch",
    "update_checkbox_group",
    "update_radio_buttons",
    "update_date",
    "update_date_range",
    "update_numeric",
    "update_select",
    "update_selectize",
    "update_slider",
    "update_task_button",
    "update_text",
    "update_text_area",
    "update_navs",
    "update_tooltip",
    "update_popover",
    # _insert
    "insert_ui",
    "remove_ui",
    # _markdown
    "markdown",
    # _modal
    "modal_button",
    "modal",
    "modal_show",
    "modal_remove",
    # _navs
    "nav_panel",
    "nav_menu",
    "nav_control",
    "nav_spacer",
    "navset_tab",
    "navset_card_tab",
    "navset_card_pill",
    "navset_card_underline",
    "navset_pill",
    "navset_underline",
    "navset_pill_list",
    "navset_hidden",
    "navset_bar",
    # _notification
    "notification_show",
    "notification_remove",
    # _output
    "output_data_frame",  # dataframe
    "output_plot",
    "output_image",
    "output_text",
    "output_code",
    "output_text_verbatim",
    "output_table",
    "output_ui",
    # _page
    "page_sidebar",
    "page_navbar",
    "page_fillable",
    "page_fluid",
    "page_fixed",
    "page_bootstrap",
    "page_auto",
    "page_output",
    # _popover
    "popover",
    # _valuebox
    "value_box",
    "value_box_theme",
    "showcase_bottom",
    "showcase_left_center",
    "showcase_top_right",
    "ValueBoxTheme",
    "ShowcaseLayout",
    # _theme
    "Theme",
    # _tooltip
    "tooltip",
    # _progress
    "Progress",
    # Items below are from htmltools
    "TagList",
    "Tag",
    "TagChild",
    "TagAttrs",
    "TagAttrValue",
    "tags",
    "HTML",
    "head_content",
    "p",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "a",
    "br",
    "div",
    "span",
    "pre",
    "code",
    "img",
    "strong",
    "em",
    "hr",
    # Submodules
    "busy_indicators",
    "fill",
    # utils
    "js_eval",
)



from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.input_radio_buttons(
        "rb",
        "Choose one:",
        {
            "html": ui.HTML("<span style='color:red;'>Red Text</span>"),
            "text": "Normal text",
        },
    ),
    ui.output_ui("val"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.ui
    def val():
        return "You chose " + input.rb()


app = App(app_ui, server)



import matplotlib.pyplot as plt
import numpy as np

from shiny.express import input, render, ui

# Equivalent to using `ui.page_fluid()` in the core API
ui.page_opts(full_width=True)

with ui.layout_sidebar():
    with ui.sidebar():
        ui.input_slider("n", "N", min=0, max=100, value=20)

    @render.plot(alt="A histogram")
    def plot() -> object:
        np.random.seed(19680801)
        x = 100 + 15 * np.random.randn(437)

        fig, ax = plt.subplots()
        ax.hist(x, input.n(), density=True)
        return fig



import asyncio
import io
import os
from datetime import date
from typing import Any

import matplotlib.pyplot as plt
import numpy as np

from shiny import App, Inputs, Outputs, Session, render, ui


def make_example(id: str, label: str, title: str, desc: str, extra: Any = None):
    return ui.column(
        4,
        ui.div(
            {"class": "card mb-4"},
            ui.div(title, class_="card-header"),
            ui.div(
                {"class": "card-body"},
                ui.p(desc, class_="card-text text-muted"),
                extra,
                ui.download_button(id, label, class_="btn-primary"),
            ),
        ),
    )


app_ui = ui.page_fluid(
    ui.row(
        make_example(
            "download1",
            label="Download CSV",
            title="Simple case",
            desc="Downloads a pre-existing file, using its existing name on disk.",
        ),
    ),
    ui.row(
        make_example(
            "download2",
            label="Download plot",
            title="Dynamic data generation",
            desc="Downloads a PNG that's generated on the fly.",
            extra=[
                ui.input_text("title", "Plot title", "Random scatter plot"),
                ui.input_slider(
                    "num_points", "Number of data points", min=1, max=100, value=50
                ),
            ],
        ),
    ),
    ui.row(
        make_example(
            "download3",
            "Download",
            "Dynamic filename",
            "Demonstrates that filenames can be generated on the fly (and use Unicode characters!).",
        ),
    ),
    ui.row(
        make_example(
            "download4",
            "Download",
            "Failed downloads",
            "Throws an error in the download handler, download should not succeed.",
        ),
    ),
    ui.row(
        make_example(
            "download5",
            "Download",
            "Undefined download",
            "This button doesn't have corresponding server code registered to it, download should result in 404 error",
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.download()
    def download1():
        """
        This is the simplest case. The implementation simply returns the name of a file.
        Note that the function name (`download1`) determines which download_button()
        corresponds to this function.
        """

        path = os.path.join(os.path.dirname(__file__), "mtcars.csv")
        return path

    @render.download(filename="image.png")
    def download2():
        """
        Another way to implement a file download is by yielding bytes; either all at
        once, like in this case, or by yielding multiple times. When using this
        approach, you should pass a filename argument to @render.download, which
        determines what the browser will name the downloaded file.
        """

        print(input.num_points())
        x = np.random.uniform(size=input.num_points())
        y = np.random.uniform(size=input.num_points())
        plt.figure()
        plt.scatter(x, y)
        plt.title(input.title())
        with io.BytesIO() as buf:
            plt.savefig(buf, format="png")
            yield buf.getvalue()

    @render.download(
        filename=lambda: f"新型-{date.today().isoformat()}-{np.random.randint(100, 999)}.csv"
    )
    async def download3():
        await asyncio.sleep(0.25)
        yield "one,two,three\n"
        yield "新,1,2\n"
        yield "型,4,5\n"

    @output(id="download4")
    @render.download(filename="failuretest.txt")
    async def _():
        yield "hello"
        raise Exception("This error was caused intentionally")


app = App(app_ui, server)
